# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Lazy setup logic for adding JAX support to treescope."""

from __future__ import annotations

import functools
import typing
from typing import Mapping

import numpy as np
from treescope import canonical_aliases
from treescope import context
from treescope import dtype_util
from treescope import lowering
from treescope import ndarray_adapters
from treescope import renderers
from treescope import rendering_parts
from treescope import repr_lib
from treescope import type_registries

# pylint: disable=import-outside-toplevel
try:
  import jax
except ImportError:
  assert not typing.TYPE_CHECKING
  jax = None
# pylint: enable=import-outside-toplevel


def _is_subdtype(dtype, base) -> bool:
  """Safely checks for dtype subtyping."""
  assert jax is not None
  jnp = jax.numpy
  try:
    return jnp.issubdtype(dtype, base)
  except TypeError:
    return False


summarization_threshold: context.ContextualValue[Mapping[str, int | None]] = (
    context.ContextualValue(
        module=__name__,
        qualname="summarization_threshold",
        initial_value={
            "tpu": 1_000_000_000,
            "gpu": 10_000_000,
            "default": 100_000,
        },
    )
)
"""Threshold for summarization of NDArrays for each backend.

This threshold determines the largest number of elements we will
summarize with summary statistics (e.g. mean, standard deviation)
when rendering in treescope. Larger values may make it slower to
display large NDArrays.

Each key should be the name of a JAX array platform, e.g. "cpu" or
"tpu". It can also be "numpy" to refer to Numpy arrays, or "default"
to refer to any other accelerator. The value is the size of the
array at which point we avoid showing summary statistics. `None`
means no limit.

This configuration argument is intended to be set at the top level
by the user, e.g. in IPython.
"""


SUMMARIZE_USING_NUMPY_THRESHOLD = 10_000_000
"""Threshold for using NumPy to summarize and render JAX Arrays.

Moving arrays to main memory and using numpy is significantly faster because
it avoid jitting the summary and rendering functions.
"""


def _is_locally_available(array: jax.Array) -> bool:
  """Checks if the array is available locally."""
  return getattr(array, "is_fully_addressable", False) or getattr(
      array, "is_fully_replicated", False
  )


def safe_to_summarize(array: jax.Array) -> bool:
  """Checks if the array is safe to summarize (not a tracer, not replicated)."""
  assert jax is not None, "JAX is not available."
  if isinstance(array, jax.core.Tracer):
    return False
  if array.is_deleted():
    return False
  if not _is_locally_available(array):
    return False
  thresh_dict = summarization_threshold.get()
  [platform] = set(device.platform for device in array.devices())
  thresh = thresh_dict.get(platform)
  if thresh is None:
    thresh = thresh_dict["default"]
  return thresh is None or array.size < thresh


def _truncate_part_with_slices(
    array: jax.Array,
    mask: jax.Array,
    prefix_slices: tuple[slice, ...],
    remaining_edge_items_per_axis: tuple[int | None, ...],
    xnp=None,
) -> tuple[jax.Array, jax.Array]:
  """Helper to truncate names of an array.

  Args:
    array: An array to truncate.
    mask: Mask array, which must be broadcastable to `array`.
    prefix_slices: Slices to apply to each axis of `array` and `mask`, starting
      at axis 0, which we have already computed.
    remaining_edge_items_per_axis: Number of edge items to keep for each axis,
      ignoring any axes whose slices are already computed in `prefix_slices`.
    xnp: backend to use (numpy or jax.numpy).

  Returns:
    Truncated array and mask, which will both be the same shape.
  """
  if xnp is None:
    assert jax is not None, "JAX is not available."
    xnp = jax.numpy

  array = xnp.array(array)
  mask = xnp.array(mask)
  mask = xnp.broadcast_to(mask, array.shape)
  if not remaining_edge_items_per_axis:
    # Perform the base case slice.
    assert len(prefix_slices) == len(array.shape)
    truncated_array = array[prefix_slices]

    valid_mask_slices = tuple(
        slice(None) if mask.shape[i] == 1 else array_slice
        for i, array_slice in enumerate(prefix_slices)
    )
    truncated_mask = xnp.broadcast_to(
        xnp.array(mask[valid_mask_slices]), truncated_array.shape
    )
    return truncated_array, truncated_mask

  # Recursive step: extract one name, run the function on each side, and
  # concatenate.
  axis = len(prefix_slices)
  edge_items = remaining_edge_items_per_axis[0]
  if edge_items is None:
    # Don't need to slice.
    return _truncate_part_with_slices(
        array,
        mask,
        prefix_slices=prefix_slices + (slice(None),),
        remaining_edge_items_per_axis=remaining_edge_items_per_axis[1:],
        xnp=xnp,
    )
  else:
    assert array.shape[axis] > 2 * edge_items
    result_a, valid_a = _truncate_part_with_slices(
        array,
        mask,
        prefix_slices=prefix_slices + (slice(None, edge_items),),
        remaining_edge_items_per_axis=remaining_edge_items_per_axis[1:],
        xnp=xnp,
    )
    result_b, valid_b = _truncate_part_with_slices(
        array,
        mask,
        prefix_slices=prefix_slices + (slice(-edge_items, None),),
        remaining_edge_items_per_axis=remaining_edge_items_per_axis[1:],
        xnp=xnp,
    )
    padding_shape = list(result_a.shape)
    padding_shape[axis] = 1
    result = xnp.concatenate(
        [result_a, xnp.zeros(padding_shape, result_a.dtype), result_b],
        axis=axis,
    )
    valid = xnp.concatenate(
        [valid_a, xnp.zeros(padding_shape, valid_a.dtype), valid_b], axis=axis
    )
    return result, valid


def truncate_array_and_mask(
    array: jax.Array,
    mask: jax.Array,
    edge_items_per_axis: tuple[int | None, ...],
) -> tuple[jax.Array, jax.Array]:
  """Truncates an array along the given axis names.

  Args:
    array: Array to truncate.
    mask: Mask array, which must have the same number of dimensions as `array`,
      and whose axis sizes must be either 1 or the same as that axis of `array`
      (e.g. they are broadcast compatible).
    edge_items_per_axis: Number of edge items to keep for each axis, ignoring
      any axes whose slices are already computed in `prefix_slices`.

  Returns:
    A tuple containing a truncated version of the array along with a valid mask.
    Values taken from the original array have the valid mask as True, and there
    is one extra element in the middle with valid as False (standing in for the
    omitted elements). The return value is always fully replicated, because
    we cannot guarantee that it is evenly sharded across devices, and this
    function is usually used immediately before copying to the host.
  """
  assert jax is not None, "JAX is not available."
  sharding_kwargs = {}
  if hasattr(array, "sharding") and hasattr(
      array.sharding, "_device_assignment"
  ):
    # _truncate_part_with_slices usually returns slices that have odd
    # dimensions, which aren't divisible by most shardings. Unfortunately,
    # the XLA GSPMD partitioner sometimes still infers a sharding over one of
    # these axes, which then leads to partitioning errors in JAX whenever we
    # try to `device_get` the resulting array or call any additional operations
    # on it. To avoid this, we'd like to tell JAX to always produce an output
    # that is not sharded over any axis. Unfortunately, this is difficult
    # because JAX requires the in_shardings and out_shardings to have the same
    # devices in the same internal order, and at the time of writing JAX does
    # not provide any public API to look up the order of the devices in a
    # sharding (it allows looking up the device *set*, but not their order).
    # Whether or not this error happens seems to be somewhat nondeterministic.
    # To avoid this, we use the private property `_device_assignment` of
    # each sharding in order to figure out what device order it has, and then
    # explicitly request a fully-replicated output that is definitely safe to
    # retrieve.
    sharding_kwargs["out_shardings"] = (
        jax.sharding.GSPMDSharding.get_replicated(
            array.sharding._device_assignment  # pylint: disable=protected-access
        )
    )
  if array.size < SUMMARIZE_USING_NUMPY_THRESHOLD and safe_to_summarize(array):
    fn = functools.partial(_truncate_part_with_slices, xnp=np)
  else:
    fn = jax.jit(
        _truncate_part_with_slices, static_argnums=(2, 3), **sharding_kwargs
    )
  return fn(array, mask, (), edge_items_per_axis)


def faster_array_repr(array: jax.Array) -> str:
  """Computes ``repr(array)``, only copying the rendered array elements.

  ``repr(array)`` on a very large jax Array can be slow, because it copies the
  entire array to host memory even when only a few elements are actually needed.
  We can avoid this by truncating the array on device before fetching it.

  Args:
    array: The array to summarize.

  Returns:
    A string representation of the array. May differ slightly from the ordinary
    ``repr``, but should contain the same elements.
  """
  assert jax is not None, "JAX is not available."
  jnp = jax.numpy
  if array.size < np.get_printoptions()["threshold"]:
    return repr(array)

  if array.aval is not None and array.aval.weak_type:
    dtype_str = f"dtype={array.dtype.name}, weak_type=True)"
  else:
    dtype_str = f"dtype={array.dtype.name})"

  edgeitems = np.get_printoptions()["edgeitems"]
  edge_items_per_axis = []
  for size in array.shape:
    if size > 2 * edgeitems + 1:
      edge_items_per_axis.append(edgeitems)
    else:
      edge_items_per_axis.append(None)
  array_edges, _ = truncate_array_and_mask(
      array,
      np.ones((1,) * array.ndim, dtype=jnp.bool_),
      edge_items_per_axis=tuple(edge_items_per_axis),
  )
  prefix = "Array("
  datastring = np.array2string(
      np.array(array_edges),
      prefix=prefix,
      suffix=",",
      separator=", ",
      threshold=0,
      edgeitems=edgeitems,
  )
  return f"{prefix}{datastring}, {dtype_str}"


def render_shape_dtype_struct(
    node: jax.ShapeDtypeStruct,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Renders jax.ShapeDtypeStruct."""
  assert jax is not None, "JAX is not available."
  if type(node) is not jax.ShapeDtypeStruct:  # pylint: disable=unidiomatic-typecheck
    return NotImplemented
  attributes = {
      "shape": node.shape,
      "dtype": node.dtype,
  }
  if node.sharding is not None:
    attributes["sharding"] = node.sharding

  # Make sure we can correctly round-trip it. We check because ShapeDtypeStruct
  # occasionally adds new attributes for new JAX features.
  rebuilt = jax.ShapeDtypeStruct(**attributes)
  if rebuilt != node:
    return NotImplemented
  else:
    return repr_lib.render_object_constructor(
        object_type=jax.ShapeDtypeStruct,
        attributes=attributes,
        path=path,
        subtree_renderer=subtree_renderer,
        roundtrippable=True,
    )


def render_precision(
    node: jax.lax.Precision,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Renders jax.lax.Precision."""
  assert jax is not None, "JAX is not available."
  if type(node) is not jax.lax.Precision:  # pylint: disable=unidiomatic-typecheck
    return NotImplemented
  return repr_lib.render_enumlike_item(
      object_type=jax.lax.Precision,
      item_name=node.name,
      item_value=node.value,
      path=path,
      subtree_renderer=subtree_renderer,
  )


def _compute_summary(
    x: jax.Array, is_floating: bool, is_integer: bool, is_bool: bool, xnp=None
) -> dict[str, jax.Array]:
  """Computes a summary of the given array."""
  if xnp is None:
    assert jax is not None, "JAX is not available."
    xnp = jax.numpy
  x = xnp.array(x)
  result = {}
  if is_floating:
    isfinite = xnp.isfinite(x)
    inf_to_nan = xnp.where(isfinite, x, xnp.array(xnp.nan, dtype=x.dtype))
    nanmean = functools.partial(xnp.nanmean, dtype=xnp.float32)
    nanstd = functools.partial(xnp.nanstd, dtype=xnp.float32)
    nanmin = lambda x: xnp.nanmin(x).astype(xnp.float32)
    nanmax = lambda x: xnp.nanmax(x).astype(xnp.float32)
    result.update(mean=nanmean(inf_to_nan), std=nanstd(inf_to_nan))
    result.update(nanmin=nanmin(x), nanmax=nanmax(x))
    result.update(
        nan=xnp.count_nonzero(xnp.isnan(x)),
        inf=xnp.count_nonzero(xnp.isposinf(x)),
    )
    result["any_finite"] = xnp.any(isfinite)
    result["-inf"] = xnp.count_nonzero(xnp.isneginf(x))
  if is_integer:
    result.update(min=xnp.min(x), max=xnp.max(x))
  if is_floating or is_integer:
    result.update(zero=xnp.count_nonzero(x == 0), nonzero=xnp.count_nonzero(x))
  if is_bool:
    result.update(
        true=xnp.count_nonzero(x), false=xnp.count_nonzero(xnp.logical_not(x))
    )
  return result


def _summarize_array_data_unconditionally(array: jax.Array) -> list[str]:
  """Summarized the data of a JAX array."""
  assert jax is not None, "JAX is not available."
  jnp = jax.numpy
  output_parts = []
  # This is required if treescope is invoked inside jitted function.
  with jax.core.ensure_compile_time_eval():
    is_floating = _is_subdtype(array.dtype, jnp.floating)
    is_integer = _is_subdtype(array.dtype, jnp.integer)
    is_bool = _is_subdtype(array.dtype, jnp.bool_)
    if array.size < SUMMARIZE_USING_NUMPY_THRESHOLD:
      compute_summary = functools.partial(_compute_summary, xnp=np)
    else:
      compute_summary = jax.jit(_compute_summary, static_argnums=(1, 2, 3))
    stat = compute_summary(array, is_floating, is_integer, is_bool)
    # Get values in parallel.
    stat = jax.device_get(stat)
    # pylint: disable=inconsistent-quotes
    if is_floating and stat["any_finite"]:
      output_parts.append(f" ≈{stat['mean']:.2} ±{stat['std']:.2}")
      output_parts.append(f" [≥{stat['nanmin']:.2}, ≤{stat['nanmax']:.2}]")

    if is_integer:
      output_parts.append(f" [≥{stat['min']:_d}, ≤{stat['max']:_d}]")
    # pylint: enable=inconsistent-quotes

    def append_if_present(output_parts, *names):
      for name in names:
        if stat[name]:
          output_parts.append(f" {name}:{stat[name]:_d}")

    if is_floating or is_integer:
      append_if_present(output_parts, "zero", "nonzero")
    if is_floating:
      append_if_present(output_parts, "nan", "inf", "-inf")

    if is_bool:
      append_if_present(output_parts, "true", "false")
    return output_parts


def summarize_array_data(array: jax.Array) -> str:
  """Summarized the data of a JAX array.

  Args:
    array: The array to summarize.

  Returns:
    A string summarizing the data of the array.
  """

  output_parts = []

  if isinstance(array, jax.core.Tracer):
    output_parts.append(" - tracer.")
  elif array.is_deleted():
    output_parts.append(" - deleted!")
  elif not _is_locally_available(array):
    output_parts.append(" - multi-host array!")
  elif safe_to_summarize(array):
    output_parts.extend(_summarize_array_data_unconditionally(array))
  else:
    output_parts.append("- too large to summarize.")

  return "".join(output_parts)


class JAXArrayAdapter(ndarray_adapters.NDArrayAdapter[jax.Array]):
  """Array adapter for JAX arrays."""

  def get_axis_info_for_array_data(
      self, array: jax.Array
  ) -> tuple[ndarray_adapters.AxisInfo, ...]:
    assert jax is not None, "JAX is not available."
    return tuple(
        ndarray_adapters.PositionalAxisInfo(i, size)
        for i, size in enumerate(array.shape)
    )

  def get_array_data_with_truncation(
      self,
      array: jax.Array,
      mask: jax.Array | None,
      edge_items_per_axis: tuple[int | None, ...],
  ) -> tuple[np.ndarray, np.ndarray]:
    assert jax is not None, "JAX is not available."
    assert not isinstance(array, jax.core.Tracer)
    assert not array.is_deleted()
    if mask is None:
      mask = np.array(True)

    if edge_items_per_axis == (None,) * array.ndim:
      # No truncation.
      return np.array(array), np.broadcast_to(mask, array.shape)

    array, mask = truncate_array_and_mask(array, mask, edge_items_per_axis)
    return jax.device_get((array, mask))

  def get_array_summary(self, array: jax.Array, fast: bool) -> str:
    output_parts = ["jax.Array "]

    output_parts.append(dtype_util.get_dtype_name(array.dtype))
    output_parts.append(repr(array.shape))
    if array.is_deleted():
      output_parts.append(" - deleted!")
    elif not fast:
      output_parts.append(summarize_array_data(array))

    return "".join(output_parts)

  def get_numpy_dtype(self, array: jax.Array) -> np.dtype | None:
    if isinstance(array.dtype, np.dtype):
      return array.dtype
    else:
      return None

  def get_sharding_info_for_array_data(
      self, array: jax.Array
  ) -> ndarray_adapters.ShardingInfo | None:
    assert jax is not None, "JAX is not available."
    if isinstance(array, jax.core.Tracer) or array.is_deleted():
      return None

    [platform] = set(device.platform for device in array.sharding.device_set)
    device_map = array.sharding.devices_indices_map(array.shape)
    return ndarray_adapters.ShardingInfo(
        shard_shape=array.sharding.shard_shape(array.shape),
        device_index_to_shard_slices={
            device.id: slices for device, slices in device_map.items()
        },
        device_type=platform.upper(),
        fully_replicated=array.is_fully_replicated,
    )

  def should_autovisualize(self, array: jax.Array) -> bool:
    assert jax is not None, "JAX is not available."
    return not isinstance(array, jax.core.Tracer) and not array.is_deleted()


def render_jax_arrays(
    node: jax.Array,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Renders a JAX array."""
  assert jax is not None, "JAX is not available."
  del subtree_renderer
  assert isinstance(node, jax.Array)
  if isinstance(node, jax.core.Tracer):
    return NotImplemented

  adapter = JAXArrayAdapter()

  if node.is_deleted():
    return rendering_parts.error_color(
        rendering_parts.text(
            "<" + adapter.get_array_summary(node, fast=True) + ">"
        )
    )

  def _placeholder() -> rendering_parts.RenderableTreePart:
    return rendering_parts.deferred_placeholder_style(
        rendering_parts.text(adapter.get_array_summary(node, fast=True))
    )

  def _thunk(placeholder_expand_state: rendering_parts.ExpandState | None):
    # Is this array simple enough to render without a summary?
    node_repr = faster_array_repr(node)
    if "\n" not in node_repr and "..." not in node_repr:
      rendering = rendering_parts.abbreviation_color(
          rendering_parts.text(f"<jax.{node_repr}>")
      )
    else:
      if node_repr.count("\n") <= 15:
        if placeholder_expand_state is None:
          default_expand_state = rendering_parts.ExpandState.WEAKLY_EXPANDED
        else:
          default_expand_state = placeholder_expand_state
      else:
        # Always start big NDArrays in collapsed mode to hide irrelevant detail.
        default_expand_state = rendering_parts.ExpandState.COLLAPSED

      # Render it with a summary.
      summarized = adapter.get_array_summary(node, fast=False)
      rendering = rendering_parts.build_custom_foldable_tree_node(
          label=rendering_parts.abbreviation_color(
              rendering_parts.comment_color_when_expanded(
                  rendering_parts.siblings(
                      rendering_parts.fold_condition(
                          expanded=rendering_parts.text("# "),
                          collapsed=rendering_parts.text("<"),
                      ),
                      summarized,
                      rendering_parts.fold_condition(
                          collapsed=rendering_parts.text(">")
                      ),
                  )
              )
          ),
          contents=rendering_parts.fold_condition(
              expanded=rendering_parts.indented_children(
                  [rendering_parts.text(node_repr)]
              )
          ),
          path=path,
          expand_state=default_expand_state,
      ).renderable

    return rendering

  return rendering_parts.RenderableAndLineAnnotations(
      renderable=lowering.maybe_defer_rendering(
          main_thunk=_thunk, placeholder_thunk=_placeholder
      ),
      annotations=rendering_parts.build_copy_button(path),
  )


def set_up_treescope():
  """Sets up treescope to render JAX objects."""
  if jax is None:
    raise RuntimeError(
        "Cannot set up JAX support in treescope: JAX cannot be imported."
    )
  type_registries.TREESCOPE_HANDLER_REGISTRY[jax.ShapeDtypeStruct] = (
      render_shape_dtype_struct
  )
  type_registries.TREESCOPE_HANDLER_REGISTRY[jax.lax.Precision] = (
      render_precision
  )

  # The concrete type of a JAX array is a private type that is dynamically
  # registered as a jax.Array subclass, so we need to add it to the list of
  # dynamically-checked virtual base classes.
  type_registries.VIRTUAL_BASE_CLASSES.append(jax.Array)
  type_registries.IMMUTABLE_TYPES_REGISTRY[jax.Array] = True
  type_registries.NDARRAY_ADAPTER_REGISTRY[jax.Array] = JAXArrayAdapter()
  type_registries.TREESCOPE_HANDLER_REGISTRY[jax.Array] = render_jax_arrays

  for jax_api_module in [
      jax.lax,
      jax.numpy,
      jax.scipy,
      jax.random,
      jax.nn,
      jax.custom_derivatives,
      jax,
  ]:
    canonical_aliases.populate_from_public_api(
        jax_api_module, canonical_aliases.prefix_filter("jax")
    )



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Lowering of Treescope's renderable parts to text or HTML.

This module defines the final logic that converts the output of node handlers
and figure parts to the final Treescope representation.
"""

from __future__ import annotations

import contextlib
import io
import json
import traceback
from typing import Any, Callable, Iterator, Sequence
import uuid

from treescope import context
from treescope import rendering_parts
from treescope._internal import html_encapsulation
from treescope._internal import html_escaping
from treescope._internal.parts import foldable_impl
from treescope._internal.parts import part_interface

_deferrables: context.ContextualValue[
    list[foldable_impl.DeferredWithThunk] | None
] = context.ContextualValue(
    module=__name__, qualname="_deferrables", initial_value=None
)
"""An optional list of accumulated deferrables, for use by this module."""


def maybe_defer_rendering(
    main_thunk: Callable[
        [rendering_parts.ExpandState | None],
        rendering_parts.RenderableTreePart,
    ],
    placeholder_thunk: Callable[[], rendering_parts.RenderableTreePart],
    expanded_newlines_for_layout: int | None = None,
) -> rendering_parts.RenderableTreePart:
  """Possibly defers rendering of a part in interactive contexts.

  This function can be used by advanced handlers and autovisualizers to delay
  the rendering of "expensive" leaves such as `jax.Array` until after the tree
  structure is drawn. If run in a non-interactive context, this just calls the
  main thunk. If run in an interactive context, it instead calls the placeholder
  thunk, and enqueues the placeholder thunk to be called later.

  Rendering can be performed in a deferred context by running the handlers under
  the `collecting_deferred_renderings` context manager, and then rendered to
  a sequence of streaming HTML updates using the `display_streaming_as_root`
  function.

  Note that handlers who call this are responsible for ensuring that the
  logic in `main_thunk` is safe to run at a later point in time. In particular,
  any rendering context managers may have been exited by the time this main
  thunk is called. As a best practice, handlers should control all of the logic
  in `main_thunk` and shouldn't recursively call the subtree renderer inside it;
  subtrees should be rendered before calling `maybe_defer_rendering`.

  Args:
    main_thunk: A callable producing the main part to render. If deferred and if
      ``expanded_newlines_for_layout`` is not None, this will be called with the
      inferred final expand state of placeholder after layout decisions.
      Otherwise, will be called with None.
    placeholder_thunk: A callable producing a placeholder object, which will be
      rendered if we are deferring rendering.
    expanded_newlines_for_layout: A guess at the extra height of this node after
      it has been expanded. If not None, this node will participate in the
      layout algorithm as if it had this height, and the final inferred
      expansion state will be passed to ``main_thunk``.

  Returns:
    Either the rendered main part or a wrapped placeholder that will later be
    replaced with the main part.
  """
  deferral_list = _deferrables.get()
  if deferral_list is None:
    return main_thunk(None)
  else:
    if expanded_newlines_for_layout is None:
      needs_layout_decision = False
      child = placeholder_thunk()
    else:
      needs_layout_decision = True
      child = foldable_impl.fake_placeholder_foldable(
          placeholder_thunk(), extra_newlines_guess=expanded_newlines_for_layout
      )
    placeholder = foldable_impl.DeferredPlaceholder(
        child=child,
        replacement_id="deferred_" + uuid.uuid4().hex,
        needs_layout_decision=needs_layout_decision,
    )
    deferral_list.append(
        foldable_impl.DeferredWithThunk(placeholder, main_thunk)
    )
    return placeholder


@contextlib.contextmanager
def collecting_deferred_renderings() -> (
    Iterator[list[foldable_impl.DeferredWithThunk]]
):
  # pylint: disable=g-doc-return-or-yield
  """Context manager that defers and collects `maybe_defer_rendering` calls.

  This context manager can be used by renderers that wish to render deferred
  objects in a streaming fashion. When used in a
  `with collecting_deferred_renderings() as deferreds:`
  expression, `deferreds` will be a list that is populated by calls to
  `maybe_defer_rendering`. This can later be passed to
  `display_streaming_as_root` to render the deferred object in a streaming
  fashion.

  Returns:
    A context manager in which `maybe_defer_rendering` calls will be deferred
    and collected into the result list.
  """
  # pylint: enable=g-doc-return-or-yield
  try:
    target = []
    with _deferrables.set_scoped(target):
      yield target
  finally:
    pass


################################################################################
# Top-level rendering and roundtrip mode implementation
################################################################################


def render_to_text_as_root(
    root_node: rendering_parts.RenderableTreePart,
    roundtrip: bool = False,
    strip_trailing_whitespace: bool = True,
    strip_whitespace_lines: bool = True,
) -> str:
  """Renders a root node to text.

  Args:
    root_node: The root node to render.
    roundtrip: Whether to render in roundtrip mode.
    strip_trailing_whitespace: Whether to remove trailing whitespace from lines.
    strip_whitespace_lines: Whether to remove lines that are entirely
      whitespace. These lines can sometimes be generated by layout code being
      conservative about line breaks. Should only be True if
      `strip_trailing_whitespace` is True.

  Returns:
    Text for the rendered node.
  """
  if strip_whitespace_lines and not strip_trailing_whitespace:
    raise ValueError(
        "strip_whitespace_lines must be False if "
        "strip_trailing_whitespace is False."
    )

  stream = io.StringIO()
  root_node.render_to_text(
      stream,
      expanded_parent=True,
      indent=0,
      roundtrip_mode=roundtrip,
      render_context={},
  )
  result = stream.getvalue()

  if strip_trailing_whitespace:
    trimmed_lines = []
    for line in result.split("\n"):
      line = line.rstrip()
      if line or not strip_whitespace_lines:
        trimmed_lines.append(line)
    result = "\n".join(trimmed_lines)
  return result


_TREESCOPE_PREAMBLE_SCRIPT = """(()=> {
  const defns = this.getRootNode().host.defns;
  let _pendingActions = [];
  let _pendingActionHandle = null;
  defns.runSoon = (work) => {
      const doWork = () => {
          const tick = performance.now();
          while (performance.now() - tick < 32) {
            if (_pendingActions.length == 0) {
                _pendingActionHandle = null;
                return;
            } else {
                const thunk = _pendingActions.shift();
                thunk();
            }
          }
          _pendingActionHandle = (
              window.requestAnimationFrame(doWork));
      };
      _pendingActions.push(work);
      if (_pendingActionHandle === null) {
          _pendingActionHandle = (
              window.requestAnimationFrame(doWork));
      }
  };
  defns.toggle_root_roundtrip = (rootelt, event) => {
      if (event.key == "r") {
          rootelt.classList.toggle("roundtrip_mode");
      }
  };
})();
"""


def _render_to_html_as_root_streaming(
    root_node: rendering_parts.RenderableTreePart,
    roundtrip: bool,
    deferreds: Sequence[foldable_impl.DeferredWithThunk],
    ignore_exceptions: bool = False,
) -> Iterator[str]:
  """Helper function: renders a root node to HTML one step at a time.

  Args:
    root_node: The root node to render.
    roundtrip: Whether to render in roundtrip mode.
    deferreds: Sequence of deferred objects to render and splice in.
    ignore_exceptions: Whether to ignore exceptions during deferred rendering,
      replacing them with error markers.

  Yields:
    HTML source for the rendered node, followed by logic to substitute each
    deferred object.
  """
  all_css_styles = set()
  all_js_defns = set()

  def _render_one(
      node,
      at_beginning_of_line: bool,
      render_context: dict[Any, Any],
      stream: io.StringIO,
  ):
    # Extract setup rules.
    setup_parts = node.html_setup_parts(foldable_impl.SETUP_CONTEXT)
    current_styles = []
    current_js_defns = []
    for part in setup_parts:
      if isinstance(part, part_interface.CSSStyleRule):
        if part not in all_css_styles:
          current_styles.append(part)
          all_css_styles.add(part)
      elif isinstance(part, part_interface.JavaScriptDefn):
        if part not in all_js_defns:
          current_js_defns.append(part)
          all_js_defns.add(part)
      else:
        raise ValueError(f"Invalid setup object: {part}")

    if current_styles:
      stream.write("<style>")
      for css_style in sorted(current_styles):
        stream.write(css_style.rule)
      stream.write("</style>")

    if current_js_defns:
      stream.write(
          "<treescope-run-here><script type='application/octet-stream'>"
      )
      for js_defn in sorted(current_js_defns):
        stream.write(js_defn.source)
      stream.write("</script></treescope-run-here>")

    # Render the node itself.
    node.render_to_html(
        stream,
        at_beginning_of_line=at_beginning_of_line,
        render_context=render_context,
    )

  # Set up the styles and scripts for the root object.
  stream = io.StringIO()
  stream.write("<style>")
  stream.write(html_escaping.without_repeated_whitespace("""
    .treescope_root {
      position: relative;
      font-family: monospace;
      white-space: pre;
      list-style-type: none;
      background-color: white;
      color: black;
      width: fit-content;
      padding-left: 2ch;
      line-height: 1.5;
      contain: content;
      content-visibility: auto;
      contain-intrinsic-size: auto none;
    }
  """))
  stream.write("</style>")
  # These scripts allow us to defer execution of javascript blocks until after
  # the content is loaded, avoiding locking up the browser rendering process.
  stream.write("<treescope-run-here><script type='application/octet-stream'>")
  stream.write(
      html_escaping.without_repeated_whitespace(_TREESCOPE_PREAMBLE_SCRIPT)
  )
  stream.write("</script></treescope-run-here>")

  # Render the root node.
  classnames = "treescope_root"
  if roundtrip:
    classnames += " roundtrip_mode"
  stream.write(
      f'<div class="{classnames}" tabindex="0" '
      'part="treescope_root" '
      'onkeydown="this.getRootNode().host.defns'
      '.toggle_root_roundtrip(this, event)">'
  )
  _render_one(root_node, True, {}, stream)
  stream.write("</div>")

  yield stream.getvalue()

  # Render any deferred parts. We insert each part into a hidden element, then
  # move them all out to their appropriate positions.
  if deferreds:
    stream = io.StringIO()
    for deferred in deferreds:
      stream.write(
          '<div style="display: none"'
          f' id="for_{deferred.placeholder.replacement_id}"><span>'
      )
      if (
          deferred.placeholder.saved_at_beginning_of_line is None
          or deferred.placeholder.saved_render_context is None
      ):
        replacement_part = rendering_parts.error_color(
            rendering_parts.text("<deferred rendering error>")
        )
      else:
        if deferred.placeholder.needs_layout_decision:
          assert isinstance(
              deferred.placeholder.child, part_interface.FoldableTreeNode
          )
          layout_decision = deferred.placeholder.child.get_expand_state()
        else:
          layout_decision = None
        try:
          replacement_part = deferred.thunk(layout_decision)
        except Exception as e:  # pylint: disable=broad-except
          if not ignore_exceptions:
            raise
          exc_child = rendering_parts.fold_condition(
              expanded=rendering_parts.indented_children(
                  [rendering_parts.text(traceback.format_exc())]
              ),
          )
          replacement_part = rendering_parts.error_color(
              rendering_parts.build_custom_foldable_tree_node(
                  label=rendering_parts.text(
                      f"<{type(e).__name__} during deferred rendering"
                  ),
                  contents=rendering_parts.siblings(
                      exc_child, rendering_parts.text(">")
                  ),
              ).renderable
          )
      _render_one(
          replacement_part,
          deferred.placeholder.saved_at_beginning_of_line,
          deferred.placeholder.saved_render_context,
          stream,
      )
      stream.write("</span></div>")

    all_ids = [deferred.placeholder.replacement_id for deferred in deferreds]
    # It's sometimes important to preserve node identity when inserting
    # deferred objects, for instance if we've already registered event listeners
    # on some nodes. However, editing the DOM in place can be slow because it
    # requires re-rendering the tree on every edit. To avoid this, we swap out
    # the tree with a clone, edit the original tree, then swap the original
    # tree back in.
    inner_script = (
        f"const targetIds = {json.dumps(all_ids)};"
        + html_escaping.without_repeated_whitespace("""
        const docroot = this.getRootNode();
        const treeroot = docroot.querySelector(".treescope_root");
        const treerootClone = treeroot.cloneNode(true);
        treeroot.replaceWith(treerootClone);
        const fragment = document.createDocumentFragment();
        fragment.appendChild(treeroot);
        for (let i = 0; i < targetIds.length; i++) {
            let target = fragment.getElementById(targetIds[i]);
            let sourceDiv = docroot.querySelector("#for_" + targetIds[i]);
            target.replaceWith(sourceDiv.firstElementChild);
            sourceDiv.remove();
        }
        treerootClone.replaceWith(treeroot);
        """)
    )
    stream.write(
        '<treescope-run-here><script type="application/octet-stream">'
        f"{inner_script}</script></treescope-run-here>"
    )
    yield stream.getvalue()


def render_to_html_as_root(
    root_node: rendering_parts.RenderableTreePart,
    roundtrip: bool = False,
    compressed: bool = False,
) -> str:
  """Renders a root node to HTML.

  This handles collecting styles and JS definitions and inserting the root
  HTML element.

  Args:
    root_node: The root node to render.
    roundtrip: Whether to render in roundtrip mode.
    compressed: Whether to compress the HTML for display.

  Returns:
    HTML source for the rendered node.
  """
  render_iterator = _render_to_html_as_root_streaming(root_node, roundtrip, [])
  html_src = "".join(render_iterator)
  return html_encapsulation.encapsulate_html(html_src, compress=compressed)


def display_streaming_as_root(
    root_node: rendering_parts.RenderableTreePart,
    deferreds: Sequence[foldable_impl.DeferredWithThunk],
    roundtrip: bool = False,
    compressed: bool = True,
    stealable: bool = False,
    ignore_exceptions: bool = False,
) -> str | None:
  """Displays a root node in an IPython notebook in a streaming fashion.

  Args:
    root_node: The root node to render.
    deferreds: Deferred objects to render and splice in.
    roundtrip: Whether to render in roundtrip mode.
    compressed: Whether to compress the HTML for display.
    stealable: Whether to return an extra HTML snippet that allows the streaming
      rendering to be relocated after it is shown.
    ignore_exceptions: Whether to ignore exceptions during deferred rendering,
      replacing them with error markers.

  Returns:
    If ``stealable`` is True, a final HTML snippet which, if inserted into a
    document, will "steal" the root node rendering, moving the DOM nodes for it
    into itself. In particular, using this as the HTML rendering of the root
    node during pretty printing will correctly associate the rendering with the
    IPython "cell output", which is visible in some IPython backends (e.g.
    JupyterLab). If ``stealable`` is False, returns None.
  """
  import IPython.display  # pylint: disable=import-outside-toplevel

  render_iterator = _render_to_html_as_root_streaming(
      root_node, roundtrip, deferreds, ignore_exceptions=ignore_exceptions
  )
  encapsulated_iterator = html_encapsulation.encapsulate_streaming_html(
      render_iterator, compress=compressed, stealable=stealable
  )

  for step in encapsulated_iterator:
    if step.segment_type == html_encapsulation.SegmentType.FINAL_OUTPUT_STEALER:
      return step.html_src
    else:
      IPython.display.display(IPython.display.HTML(step.html_src))



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for treescope's internal foldable representation."""

import dataclasses
import io
from typing import Any, Sequence

from absl.testing import absltest
from absl.testing import parameterized
import jax
from treescope import lowering
from treescope import type_registries
from treescope._internal.parts import basic_parts
from treescope._internal.parts import common_structures
from treescope._internal.parts import part_interface

CSSStyleRule = part_interface.CSSStyleRule


@dataclasses.dataclass(frozen=True)
class MockFoldableTreeNode(part_interface.FoldableTreeNode):
  tag: str

  def get_expand_state(self):
    raise NotImplementedError()

  def set_expand_state(self, value):
    pass

  def as_expanded_part(self):
    raise NotImplementedError()

  def html_setup_parts(self, *args, **kwargs):
    raise NotImplementedError()

  def render_to_html(self, *args, **kwargs):
    raise NotImplementedError()

  def render_to_text(self, *args, **kwargs):
    raise NotImplementedError()


@dataclasses.dataclass(frozen=True)
class MockRenderableTreePart(part_interface.RenderableTreePart):
  tag: StopAsyncIteration

  def _compute_collapsed_width(self) -> int:
    return 100

  def _compute_newlines_in_expanded_parent(self) -> int:
    return 10

  def _compute_layout_marks_in_this_part(self) -> frozenset[Any]:
    return frozenset([self.tag])

  def foldables_in_this_part(self):
    return (MockFoldableTreeNode(self.tag),)

  def html_setup_parts(self, context):
    return {part_interface.CSSStyleRule(f'[mock rule {self.tag}]')}

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    b = 'y' if at_beginning_of_line else 'n'
    stream.write(f'[mock {self.tag}, b:{b}]')

  def render_to_text(
      self,
      stream: io.TextIOBase,
      *,
      expanded_parent: bool,
      indent: int,
      roundtrip_mode: bool,
      render_context: dict[Any, Any],
  ):
    e = 'y' if expanded_parent else 'n'
    r = 'y' if roundtrip_mode else 'n'
    stream.write(f'[mock {self.tag}, e:{e} r:{r}]')


mock_context = part_interface.HtmlContextForSetup(
    collapsed_selector='[mock collapsed_selector]',
    roundtrip_selector='[mock roundtrip_selector]',
)


class SpanGroupForTest_OneRule(basic_parts.BaseSpanGroup):  # pylint: disable=invalid-name
  """Test subclass of BaseSpanGroup."""

  def _span_css_class(self) -> str:
    return 'test_span_css_class'

  def _span_css_rule(self, context):
    return CSSStyleRule('test_span_css_rule')


class SpanGroupForTest_TwoRules(basic_parts.BaseSpanGroup):  # pylint: disable=invalid-name
  """Test subclass of BaseSpanGroup."""

  def _span_css_class(self) -> str:
    return 'test_span_css_class'

  def _span_css_rule(self, context):
    return {
        CSSStyleRule('test_span_css_rule'),
        CSSStyleRule('test_span_css_rule_2'),
    }


class RepresentationPartsTest(parameterized.TestCase):

  @parameterized.named_parameters(
      dict(
          testcase_name='empty_part',
          part=basic_parts.EmptyPart(),
          expected_width=0,
          expected_newlines=0,
          expected_tags=frozenset({}),
          expected_foldables=(),
          expected_setup_parts=set(),
          expected_text_collapsed='',
          expected_text_expanded='',
          expected_text_roundtrip_collapsed='',
          expected_text_roundtrip_expanded='',
          expected_html='',
          expected_html_at_beginning='',
      ),
      dict(
          testcase_name='single_line_text',
          part=basic_parts.Text('some text with characters < > &'),
          expected_width=31,
          expected_newlines=0,
          expected_tags=frozenset({}),
          expected_foldables=(),
          expected_setup_parts=set(),
          expected_text_collapsed='some text with characters < > &',
          expected_text_expanded='some text with characters < > &',
          expected_text_roundtrip_collapsed='some text with characters < > &',
          expected_text_roundtrip_expanded='some text with characters < > &',
          expected_html='some text with characters &lt; &gt; &amp;',
          expected_html_at_beginning=(
              'some text with characters &lt; &gt; &amp;'
          ),
      ),
      dict(
          testcase_name='siblings',
          part=basic_parts.Siblings([
              MockRenderableTreePart('A'),
              MockRenderableTreePart('B'),
              MockRenderableTreePart('C'),
          ]),
          expected_width=300,
          expected_newlines=30,
          expected_tags=frozenset({'C', 'B', 'A'}),
          expected_foldables=[
              MockFoldableTreeNode(tag='A'),
              MockFoldableTreeNode(tag='B'),
              MockFoldableTreeNode(tag='C'),
          ],
          expected_setup_parts={
              CSSStyleRule(rule='[mock rule A]'),
              CSSStyleRule(rule='[mock rule B]'),
              CSSStyleRule(rule='[mock rule C]'),
          },
          expected_text_collapsed=(
              '[mock A, e:n r:n][mock B, e:n r:n][mock C, e:n r:n]'
          ),
          expected_text_expanded=(
              '[mock A, e:y r:n][mock B, e:y r:n][mock C, e:y r:n]'
          ),
          expected_text_roundtrip_collapsed=(
              '[mock A, e:n r:y][mock B, e:n r:y][mock C, e:n r:y]'
          ),
          expected_text_roundtrip_expanded=(
              '[mock A, e:y r:y][mock B, e:y r:y][mock C, e:y r:y]'
          ),
          expected_html='[mock A, b:n][mock B, b:n][mock C, b:n]',
          expected_html_at_beginning='[mock A, b:y][mock B, b:n][mock C, b:n]',
      ),
      dict(
          testcase_name='tagged_group',
          part=basic_parts.WithLayoutMark(
              MockRenderableTreePart('A'), 'test_tag'
          ),
          expected_width=100,
          expected_newlines=10,
          expected_tags=frozenset({'test_tag', 'A'}),
          expected_foldables=(MockFoldableTreeNode(tag='A'),),
          expected_setup_parts={CSSStyleRule(rule='[mock rule A]')},
          expected_text_collapsed='[mock A, e:n r:n]',
          expected_text_expanded='[mock A, e:y r:n]',
          expected_text_roundtrip_collapsed='[mock A, e:n r:y]',
          expected_text_roundtrip_expanded='[mock A, e:y r:y]',
          expected_html='[mock A, b:n]',
          expected_html_at_beginning='[mock A, b:y]',
      ),
      dict(
          testcase_name='span_group_1',
          part=SpanGroupForTest_OneRule(MockRenderableTreePart('A')),
          expected_width=100,
          expected_newlines=10,
          expected_tags=frozenset({'A'}),
          expected_foldables=(MockFoldableTreeNode(tag='A'),),
          expected_setup_parts={
              CSSStyleRule(rule='test_span_css_rule'),
              CSSStyleRule(rule='[mock rule A]'),
          },
          expected_text_collapsed='[mock A, e:n r:n]',
          expected_text_expanded='[mock A, e:y r:n]',
          expected_text_roundtrip_collapsed='[mock A, e:n r:y]',
          expected_text_roundtrip_expanded='[mock A, e:y r:y]',
          expected_html=(
              '<span class="test_span_css_class">[mock A, b:n]</span>'
          ),
          expected_html_at_beginning=(
              '<span class="test_span_css_class">[mock A, b:y]</span>'
          ),
      ),
      dict(
          testcase_name='span_group_2',
          part=SpanGroupForTest_TwoRules(MockRenderableTreePart('A')),
          expected_width=100,
          expected_newlines=10,
          expected_tags=frozenset({'A'}),
          expected_foldables=(MockFoldableTreeNode(tag='A'),),
          expected_setup_parts={
              CSSStyleRule(rule='test_span_css_rule'),
              CSSStyleRule(rule='test_span_css_rule_2'),
              CSSStyleRule(rule='[mock rule A]'),
          },
          expected_text_collapsed='[mock A, e:n r:n]',
          expected_text_expanded='[mock A, e:y r:n]',
          expected_text_roundtrip_collapsed='[mock A, e:n r:y]',
          expected_text_roundtrip_expanded='[mock A, e:y r:y]',
          expected_html=(
              '<span class="test_span_css_class">[mock A, b:n]</span>'
          ),
          expected_html_at_beginning=(
              '<span class="test_span_css_class">[mock A, b:y]</span>'
          ),
      ),
      dict(
          testcase_name='fold_condition_collapsed',
          part=basic_parts.FoldCondition(
              collapsed=MockRenderableTreePart(tag='A')
          ),
          expected_width=100,
          expected_newlines=0,
          expected_tags=frozenset({'A'}),
          expected_foldables=(),
          expected_setup_parts={CSSStyleRule(rule='[mock rule A]'), ...},
          expected_text_collapsed='[mock A, e:n r:n]',
          expected_text_expanded='',
          expected_text_roundtrip_collapsed='[mock A, e:n r:y]',
          expected_text_roundtrip_expanded='',
          expected_html='<span class="when_collapsed">[mock A, b:n]</span>',
          expected_html_at_beginning=(
              '<span class="when_collapsed">[mock A, b:y]</span>'
          ),
      ),
      dict(
          testcase_name='fold_condition_expanded',
          part=basic_parts.FoldCondition(
              expanded=MockRenderableTreePart(tag='B')
          ),
          expected_width=0,
          expected_newlines=10,
          expected_tags=frozenset({'B'}),
          expected_foldables=(MockFoldableTreeNode(tag='B'),),
          expected_setup_parts={CSSStyleRule(rule='[mock rule B]'), ...},
          expected_text_collapsed='',
          expected_text_expanded='[mock B, e:y r:n]',
          expected_text_roundtrip_collapsed='',
          expected_text_roundtrip_expanded='[mock B, e:y r:y]',
          expected_html='<span class="when_expanded">[mock B, b:n]</span>',
          expected_html_at_beginning=(
              '<span class="when_expanded">[mock B, b:y]</span>'
          ),
      ),
      dict(
          testcase_name='fold_condition_both',
          part=basic_parts.FoldCondition(
              expanded=MockRenderableTreePart(tag='A'),
              collapsed=MockRenderableTreePart(tag='B'),
          ),
          expected_width=100,
          expected_newlines=10,
          expected_tags=frozenset({'B', 'A'}),
          expected_foldables=(MockFoldableTreeNode(tag='A'),),
          expected_setup_parts={
              CSSStyleRule(rule='[mock rule A]'),
              CSSStyleRule(rule='[mock rule B]'),
              ...,
          },
          expected_text_collapsed='[mock B, e:n r:n]',
          expected_text_expanded='[mock A, e:y r:n]',
          expected_text_roundtrip_collapsed='[mock B, e:n r:y]',
          expected_text_roundtrip_expanded='[mock A, e:y r:y]',
          expected_html=(
              '<span class="when_collapsed">[mock B, b:n]</span><span'
              ' class="when_expanded">[mock A, b:n]</span>'
          ),
          expected_html_at_beginning=(
              '<span class="when_collapsed">[mock B, b:y]</span><span'
              ' class="when_expanded">[mock A, b:y]</span>'
          ),
      ),
      dict(
          testcase_name='roundtrip_condition_roundtrip',
          part=basic_parts.RoundtripCondition(
              roundtrip=MockRenderableTreePart(tag='A')
          ),
          expected_width=0,
          expected_newlines=0,
          expected_tags=frozenset({'A'}),
          expected_foldables=(),
          expected_setup_parts={CSSStyleRule(rule='[mock rule A]'), ...},
          expected_text_collapsed='',
          expected_text_expanded='',
          expected_text_roundtrip_collapsed='[mock A, e:n r:y]',
          expected_text_roundtrip_expanded='[mock A, e:y r:y]',
          expected_html='<span class="when_roundtrip">[mock A, b:n]</span>',
          expected_html_at_beginning=(
              '<span class="when_roundtrip">[mock A, b:y]</span>'
          ),
      ),
      dict(
          testcase_name='roundtrip_condition_not_roundtrip',
          part=basic_parts.RoundtripCondition(
              not_roundtrip=MockRenderableTreePart(tag='B')
          ),
          expected_width=100,
          expected_newlines=10,
          expected_tags=frozenset({'B'}),
          expected_foldables=(MockFoldableTreeNode(tag='B'),),
          expected_setup_parts={CSSStyleRule(rule='[mock rule B]'), ...},
          expected_text_collapsed='[mock B, e:n r:n]',
          expected_text_expanded='[mock B, e:y r:n]',
          expected_text_roundtrip_collapsed='',
          expected_text_roundtrip_expanded='',
          expected_html='<span class="when_not_roundtrip">[mock B, b:n]</span>',
          expected_html_at_beginning=(
              '<span class="when_not_roundtrip">[mock B, b:y]</span>'
          ),
      ),
      dict(
          testcase_name='roundtrip_condition_both',
          part=basic_parts.RoundtripCondition(
              roundtrip=MockRenderableTreePart(tag='A'),
              not_roundtrip=MockRenderableTreePart(tag='B'),
          ),
          expected_width=100,
          expected_newlines=10,
          expected_tags=frozenset({'B', 'A'}),
          expected_foldables=(MockFoldableTreeNode(tag='B'),),
          expected_setup_parts={
              CSSStyleRule(rule='[mock rule A]'),
              CSSStyleRule(rule='[mock rule B]'),
              ...,
          },
          expected_text_collapsed='[mock B, e:n r:n]',
          expected_text_expanded='[mock B, e:y r:n]',
          expected_text_roundtrip_collapsed='[mock A, e:n r:y]',
          expected_text_roundtrip_expanded='[mock A, e:y r:y]',
          expected_html=(
              '<span class="when_roundtrip">[mock A, b:n]</span><span'
              ' class="when_not_roundtrip">[mock B, b:n]</span>'
          ),
          expected_html_at_beginning=(
              '<span class="when_roundtrip">[mock A, b:y]</span><span'
              ' class="when_not_roundtrip">[mock B, b:y]</span>'
          ),
      ),
      dict(
          testcase_name='summarizable_condition_summary',
          part=basic_parts.SummarizableCondition(
              summary=MockRenderableTreePart(tag='A')
          ),
          expected_width=100,
          expected_newlines=0,
          expected_tags=frozenset({'A'}),
          expected_foldables=(),
          expected_setup_parts={CSSStyleRule(rule='[mock rule A]'), ...},
          expected_text_collapsed='[mock A, e:n r:n]',
          expected_text_expanded='',
          expected_text_roundtrip_collapsed='',
          expected_text_roundtrip_expanded='',
          expected_html=(
              '<span class="when_collapsed_and_not_roundtrip">[mock A,'
              ' b:n]</span>'
          ),
          expected_html_at_beginning=(
              '<span class="when_collapsed_and_not_roundtrip">[mock A,'
              ' b:y]</span>'
          ),
      ),
      dict(
          testcase_name='summarizable_condition_detail',
          part=basic_parts.SummarizableCondition(
              detail=MockRenderableTreePart(tag='B')
          ),
          expected_width=0,
          expected_newlines=10,
          expected_tags=frozenset({'B'}),
          expected_foldables=(MockFoldableTreeNode(tag='B'),),
          expected_setup_parts={CSSStyleRule(rule='[mock rule B]'), ...},
          expected_text_collapsed='',
          expected_text_expanded='[mock B, e:y r:n]',
          expected_text_roundtrip_collapsed='[mock B, e:n r:y]',
          expected_text_roundtrip_expanded='[mock B, e:y r:y]',
          expected_html=(
              '<span class="when_expanded_or_roundtrip">[mock B, b:n]</span>'
          ),
          expected_html_at_beginning=(
              '<span class="when_expanded_or_roundtrip">[mock B, b:y]</span>'
          ),
      ),
      dict(
          testcase_name='summarizable_condition_both',
          part=basic_parts.SummarizableCondition(
              summary=MockRenderableTreePart(tag='A'),
              detail=MockRenderableTreePart(tag='B'),
          ),
          expected_width=100,
          expected_newlines=10,
          expected_tags=frozenset({'B', 'A'}),
          expected_foldables=(MockFoldableTreeNode(tag='B'),),
          expected_setup_parts={
              CSSStyleRule(rule='[mock rule A]'),
              CSSStyleRule(rule='[mock rule B]'),
              ...,
          },
          expected_text_collapsed='[mock A, e:n r:n]',
          expected_text_expanded='[mock B, e:y r:n]',
          expected_text_roundtrip_collapsed='[mock B, e:n r:y]',
          expected_text_roundtrip_expanded='[mock B, e:y r:y]',
          expected_html=(
              '<span class="when_collapsed_and_not_roundtrip">[mock A,'
              ' b:n]</span><span class="when_expanded_or_roundtrip">[mock B,'
              ' b:n]</span>'
          ),
          expected_html_at_beginning=(
              '<span class="when_collapsed_and_not_roundtrip">[mock A,'
              ' b:y]</span><span class="when_expanded_or_roundtrip">[mock B,'
              ' b:y]</span>'
          ),
      ),
      dict(
          testcase_name='on_separate_lines',
          part=basic_parts.OnSeparateLines([
              MockRenderableTreePart('A'),
              MockRenderableTreePart('B'),
              MockRenderableTreePart('C'),
          ]),
          indent=4,
          expected_width=300,
          expected_newlines=34,
          expected_tags=frozenset({'C', 'B', 'A'}),
          expected_foldables=[
              MockFoldableTreeNode(tag='A'),
              MockFoldableTreeNode(tag='B'),
              MockFoldableTreeNode(tag='C'),
          ],
          expected_setup_parts={
              CSSStyleRule(rule='[mock rule A]'),
              CSSStyleRule(rule='[mock rule B]'),
              CSSStyleRule(rule='[mock rule C]'),
              ...,
          },
          expected_text_collapsed=(
              '[mock A, e:n r:n][mock B, e:n r:n][mock C, e:n r:n]'
          ),
          expected_text_expanded=(
              '\n    [mock A, e:y r:n]'
              '\n    [mock B, e:y r:n]'
              '\n    [mock C, e:y r:n]'
              '\n    '
          ),
          expected_text_roundtrip_collapsed=(
              '[mock A, e:n r:y][mock B, e:n r:y][mock C, e:n r:y]'
          ),
          expected_text_roundtrip_expanded=(
              '\n    [mock A, e:y r:y]'
              '\n    [mock B, e:y r:y]'
              '\n    [mock C, e:y r:y]'
              '\n    '
          ),
          expected_html=(
              '<span class="separate_lines_children"><span'
              ' class="separate_lines_child">[mock A, b:y]</span><span'
              ' class="separate_lines_child">[mock B, b:y]</span><span'
              ' class="separate_lines_child">[mock C, b:y]</span></span>'
          ),
          expected_html_at_beginning=(
              '<span class="separate_lines_children"><span'
              ' class="separate_lines_child">[mock A, b:y]</span><span'
              ' class="separate_lines_child">[mock B, b:y]</span><span'
              ' class="separate_lines_child">[mock C, b:y]</span></span>'
          ),
      ),
      dict(
          testcase_name='indented_children',
          part=basic_parts.IndentedChildren([
              MockRenderableTreePart('A'),
              MockRenderableTreePart('B'),
              MockRenderableTreePart('C'),
          ]),
          indent=4,
          expected_width=300,
          expected_newlines=34,
          expected_tags=frozenset({'C', 'B', 'A'}),
          expected_foldables=[
              MockFoldableTreeNode(tag='A'),
              MockFoldableTreeNode(tag='B'),
              MockFoldableTreeNode(tag='C'),
          ],
          expected_setup_parts={
              CSSStyleRule(rule='[mock rule A]'),
              CSSStyleRule(rule='[mock rule B]'),
              CSSStyleRule(rule='[mock rule C]'),
              ...,
          },
          expected_text_collapsed=(
              '[mock A, e:n r:n][mock B, e:n r:n][mock C, e:n r:n]'
          ),
          expected_text_expanded=(
              '\n      [mock A, e:y r:n]'
              '\n      [mock B, e:y r:n]'
              '\n      [mock C, e:y r:n]'
              '\n    '
          ),
          expected_text_roundtrip_collapsed=(
              '[mock A, e:n r:y][mock B, e:n r:y][mock C, e:n r:y]'
          ),
          expected_text_roundtrip_expanded=(
              '\n      [mock A, e:y r:y]'
              '\n      [mock B, e:y r:y]'
              '\n      [mock C, e:y r:y]'
              '\n    '
          ),
          expected_html=(
              '<span class="indented_children"><span'
              ' class="indented_child">[mock A, b:y]</span><span'
              ' class="indented_child">[mock B, b:y]</span><span'
              ' class="indented_child">[mock C, b:y]</span></span>'
          ),
          expected_html_at_beginning=(
              '<span class="indented_children"><span'
              ' class="indented_child">[mock A, b:y]</span><span'
              ' class="indented_child">[mock B, b:y]</span><span'
              ' class="indented_child">[mock C, b:y]</span></span>'
          ),
      ),
      dict(
          testcase_name='box_with_outline',
          part=basic_parts.StyledBoxWithOutline(
              MockRenderableTreePart('A'), css_style='--foo: "1";'
          ),
          indent=4,
          expected_width=100,
          expected_newlines=12,
          expected_tags=frozenset({'A'}),
          expected_foldables=(MockFoldableTreeNode(tag='A'),),
          expected_setup_parts={
              CSSStyleRule(rule='[mock rule A]'),
              ...,
          },
          expected_text_collapsed='[mock A, e:n r:n]',
          expected_text_expanded=(
              '\n    #╭┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄╮'
              '\n'
              '\n    [mock A, e:y r:n]'
              '\n    #╰┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄╯'
              '\n    '
          ),
          expected_text_roundtrip_collapsed='[mock A, e:n r:y]',
          expected_text_roundtrip_expanded=(
              '\n    #╭┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄╮'
              '\n'
              '\n    [mock A, e:y r:y]'
              '\n    #╰┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄╯'
              '\n    '
          ),
          expected_html=(
              '<span class="outerbox_for_outline"><span'
              ' class="box_with_outline" style="--foo: &quot;1&quot;;">[mock A,'
              ' b:y]</span></span>'
          ),
          expected_html_at_beginning=(
              '<span class="outerbox_for_outline"><span'
              ' class="box_with_outline" style="--foo: &quot;1&quot;;">[mock A,'
              ' b:y]</span></span>'
          ),
      ),
      dict(
          testcase_name='with_hover_tooltip',
          part=basic_parts.WithHoverTooltip(
              MockRenderableTreePart('A'), tooltip='test_tooltip'
          ),
          indent=4,
          expected_width=100,
          expected_newlines=10,
          expected_tags=frozenset({'A'}),
          expected_foldables=(MockFoldableTreeNode(tag='A'),),
          expected_setup_parts={
              CSSStyleRule(rule='[mock rule A]'),
              ...,
          },
          expected_text_collapsed='[mock A, e:n r:n]',
          expected_text_expanded='[mock A, e:y r:n]',
          expected_text_roundtrip_collapsed='[mock A, e:n r:y]',
          expected_text_roundtrip_expanded='[mock A, e:y r:y]',
          expected_html=(
              '<span class="has_hover_tooltip"'
              ' data-tooltip="test_tooltip">[mock A, b:n]</span>'
          ),
          expected_html_at_beginning=(
              '<span class="has_hover_tooltip"'
              ' data-tooltip="test_tooltip">[mock A, b:y]</span>'
          ),
      ),
      dict(
          testcase_name='scoped_unselectable',
          part=basic_parts.ScopedSelectableAnnotation(
              MockRenderableTreePart('A')
          ),
          indent=4,
          expected_width=100,
          expected_newlines=10,
          expected_tags=frozenset({'A'}),
          expected_foldables=(MockFoldableTreeNode(tag='A'),),
          expected_setup_parts={
              CSSStyleRule(rule='[mock rule A]'),
              ...,
          },
          expected_text_collapsed='[mock A, e:n r:n]',
          expected_text_expanded='[mock A, e:y r:n]',
          expected_text_roundtrip_collapsed='',
          expected_text_roundtrip_expanded='\n    # [mock A, e:y r:y]\n    ',
          expected_html=(
              '<span tabindex="-1" class="scoped_unselectable">[mock A,'
              ' b:n]</span>'
          ),
          expected_html_at_beginning=(
              '<span tabindex="-1" class="scoped_unselectable">[mock A,'
              ' b:y]</span>'
          ),
      ),
  )
  def test_parts(
      self,
      *,
      part: basic_parts.RenderableTreePart,
      expected_width: int | type(NotImplemented),
      expected_newlines: int,
      expected_tags: set[Any],
      expected_foldables: Sequence[basic_parts.FoldableTreeNode],
      expected_setup_parts: set[Any],
      expected_text_collapsed: str,
      expected_text_expanded: str,
      expected_text_roundtrip_collapsed: str,
      expected_text_roundtrip_expanded: str,
      expected_html: str,
      expected_html_at_beginning: str,
      indent: int = 0,
  ):
    with self.subTest('collapsed_width'):
      if expected_width is not NotImplemented:
        self.assertEqual(part.collapsed_width, expected_width)

    with self.subTest('newlines_in_expanded_parent'):
      self.assertEqual(part.newlines_in_expanded_parent, expected_newlines)

    with self.subTest('layout_marks_in_this_part'):
      self.assertEqual(part.layout_marks_in_this_part, expected_tags)

    with self.subTest('foldables_in_this_part'):
      self.assertEqual(part.foldables_in_this_part(), expected_foldables)

    with self.subTest('setup_parts'):
      if isinstance(expected_setup_parts, set) and ... in expected_setup_parts:
        self.assertContainsSubset(
            expected_setup_parts - {...}, part.html_setup_parts(mock_context)
        )
      else:
        self.assertEqual(
            expected_setup_parts, part.html_setup_parts(mock_context)
        )

    # Check rendering in text mode.
    with self.subTest('text_collapsed'):
      stream = io.StringIO()
      part.render_to_text(
          stream,
          expanded_parent=False,
          indent=indent,
          roundtrip_mode=False,
          render_context={},
      )
      self.assertEqual(stream.getvalue(), expected_text_collapsed)

    with self.subTest('text_expanded'):
      stream = io.StringIO()
      part.render_to_text(
          stream,
          expanded_parent=True,
          indent=indent,
          roundtrip_mode=False,
          render_context={},
      )
      self.assertEqual(stream.getvalue(), expected_text_expanded)

    with self.subTest('text_roundtrip_collapsed'):
      stream = io.StringIO()
      part.render_to_text(
          stream,
          expanded_parent=False,
          indent=indent,
          roundtrip_mode=True,
          render_context={},
      )
      self.assertEqual(stream.getvalue(), expected_text_roundtrip_collapsed)

    with self.subTest('text_roundtrip_expanded'):
      stream = io.StringIO()
      part.render_to_text(
          stream,
          expanded_parent=True,
          indent=indent,
          roundtrip_mode=True,
          render_context={},
      )
      self.assertEqual(stream.getvalue(), expected_text_roundtrip_expanded)

    with self.subTest('html'):
      stream = io.StringIO()
      part.render_to_html(stream, at_beginning_of_line=False, render_context={})
      self.assertEqual(stream.getvalue(), expected_html)

    with self.subTest('html_at_beginning'):
      stream = io.StringIO()
      part.render_to_html(stream, at_beginning_of_line=True, render_context={})
      self.assertEqual(stream.getvalue(), expected_html_at_beginning)

  def test_multiline_text_part(self):
    part = basic_parts.Text('some text\nsecond line\nthird line')
    self.assertEqual(part.newlines_in_expanded_parent, 2)

    stream = io.StringIO()
    part.render_to_text(
        stream,
        expanded_parent=True,
        indent=4,
        roundtrip_mode=False,
        render_context={},
    )
    self.assertEqual(
        stream.getvalue(),
        'some text\n    second line\n    third line',
    )

  def test_build_siblings(self):
    part = basic_parts.siblings(
        'foo',
        MockRenderableTreePart(tag='A'),
        'bar',
        basic_parts.Siblings(
            [MockRenderableTreePart('B'), MockRenderableTreePart('C')],
        ),
        basic_parts.EmptyPart(),
        'baz',
    )
    self.assertEqual(
        part,
        basic_parts.Siblings(
            children=(
                basic_parts.Text(text='foo'),
                MockRenderableTreePart(tag='A'),
                basic_parts.Text(text='bar'),
                MockRenderableTreePart(tag='B'),
                MockRenderableTreePart(tag='C'),
                basic_parts.Text(text='baz'),
            )
        ),
    )

  def test_build_siblings_with_annotations(self):
    part = basic_parts.siblings_with_annotations(
        'foo',
        basic_parts.RenderableAndLineAnnotations(
            MockRenderableTreePart(tag='A'),
            MockRenderableTreePart(tag='annotation1'),
        ),
        'bar',
        basic_parts.Siblings(
            [MockRenderableTreePart('B'), MockRenderableTreePart('C')],
        ),
        basic_parts.RenderableAndLineAnnotations(
            MockRenderableTreePart(tag='D'),
            MockRenderableTreePart(tag='annotation2'),
        ),
        basic_parts.EmptyPart(),
        'baz',
        extra_annotations=[
            MockRenderableTreePart(tag='annotation3'),
        ],
    )
    self.assertEqual(
        part,
        basic_parts.RenderableAndLineAnnotations(
            renderable=basic_parts.Siblings(
                children=(
                    basic_parts.Text(text='foo'),
                    MockRenderableTreePart(tag='A'),
                    basic_parts.Text(text='bar'),
                    MockRenderableTreePart(tag='B'),
                    MockRenderableTreePart(tag='C'),
                    MockRenderableTreePart(tag='D'),
                    basic_parts.Text(text='baz'),
                )
            ),
            annotations=basic_parts.Siblings(
                children=(
                    MockRenderableTreePart(tag='annotation1'),
                    MockRenderableTreePart(tag='annotation2'),
                    MockRenderableTreePart(tag='annotation3'),
                )
            ),
        ),
    )

  def test_build_full_line_with_annotations(self):
    part = basic_parts.build_full_line_with_annotations(
        'foo',
        basic_parts.RenderableAndLineAnnotations(
            MockRenderableTreePart(tag='A'),
            MockRenderableTreePart(tag='annotation1'),
        ),
        'bar',
        basic_parts.RenderableAndLineAnnotations(
            MockRenderableTreePart(tag='B'),
            MockRenderableTreePart(tag='annotation2'),
        ),
    )
    self.assertEqual(
        part,
        basic_parts.Siblings(
            children=(
                basic_parts.Text(text='foo'),
                MockRenderableTreePart(tag='A'),
                basic_parts.Text(text='bar'),
                MockRenderableTreePart(tag='B'),
                basic_parts.FoldCondition(
                    collapsed=basic_parts.EmptyPart(),
                    expanded=basic_parts.Siblings(
                        children=(
                            MockRenderableTreePart(tag='annotation1'),
                            MockRenderableTreePart(tag='annotation2'),
                        )
                    ),
                ),
            )
        ),
    )

  def test_build_children(self):
    children = [
        'foo',
        basic_parts.RenderableAndLineAnnotations(
            MockRenderableTreePart(tag='A'),
            MockRenderableTreePart(tag='annotation2'),
        ),
        'bar',
        MockRenderableTreePart(tag='B'),
    ]
    with self.subTest('OnSeparateLines'):
      self.assertEqual(
          basic_parts.OnSeparateLines.build(children),
          basic_parts.OnSeparateLines(
              children=[
                  basic_parts.Siblings(
                      children=(basic_parts.Text(text='foo'),)
                  ),
                  basic_parts.Siblings(
                      children=(
                          MockRenderableTreePart(tag='A'),
                          basic_parts.FoldCondition(
                              collapsed=basic_parts.EmptyPart(),
                              expanded=basic_parts.Siblings(
                                  children=(
                                      MockRenderableTreePart(tag='annotation2'),
                                  )
                              ),
                          ),
                      )
                  ),
                  basic_parts.Siblings(
                      children=(basic_parts.Text(text='bar'),)
                  ),
                  basic_parts.Siblings(
                      children=(MockRenderableTreePart(tag='B'),)
                  ),
              ]
          ),
      )
    with self.subTest('IndentedChildren'):
      self.assertEqual(
          basic_parts.IndentedChildren.build(children),
          basic_parts.IndentedChildren(
              children=[
                  basic_parts.Siblings(
                      children=(basic_parts.Text(text='foo'),)
                  ),
                  basic_parts.Siblings(
                      children=(
                          MockRenderableTreePart(tag='A'),
                          basic_parts.FoldCondition(
                              collapsed=basic_parts.EmptyPart(),
                              expanded=basic_parts.Siblings(
                                  children=(
                                      MockRenderableTreePart(tag='annotation2'),
                                  )
                              ),
                          ),
                      )
                  ),
                  basic_parts.Siblings(
                      children=(basic_parts.Text(text='bar'),)
                  ),
                  basic_parts.Siblings(
                      children=(MockRenderableTreePart(tag='B'),)
                  ),
              ]
          ),
      )
    with self.subTest('IndentedChildren_comma_separated'):
      self.assertEqual(
          basic_parts.IndentedChildren.build(children, comma_separated=True),
          basic_parts.IndentedChildren(
              children=[
                  basic_parts.Siblings(
                      children=(
                          basic_parts.Text(text='foo'),
                          basic_parts.Text(text=','),
                          basic_parts.FoldCondition(
                              collapsed=basic_parts.Text(text=' '),
                              expanded=basic_parts.EmptyPart(),
                          ),
                      )
                  ),
                  basic_parts.Siblings(
                      children=(
                          MockRenderableTreePart(tag='A'),
                          basic_parts.Text(text=','),
                          basic_parts.FoldCondition(
                              collapsed=basic_parts.Text(text=' '),
                              expanded=basic_parts.EmptyPart(),
                          ),
                          basic_parts.FoldCondition(
                              collapsed=basic_parts.EmptyPart(),
                              expanded=basic_parts.Siblings(
                                  children=(
                                      MockRenderableTreePart(tag='annotation2'),
                                  )
                              ),
                          ),
                      )
                  ),
                  basic_parts.Siblings(
                      children=(
                          basic_parts.Text(text='bar'),
                          basic_parts.Text(text=','),
                          basic_parts.FoldCondition(
                              collapsed=basic_parts.Text(text=' '),
                              expanded=basic_parts.EmptyPart(),
                          ),
                      )
                  ),
                  basic_parts.Siblings(
                      children=(
                          MockRenderableTreePart(tag='B'),
                          basic_parts.FoldCondition(
                              collapsed=basic_parts.EmptyPart(),
                              expanded=basic_parts.Text(text=','),
                          ),
                      )
                  ),
              ]
          ),
      )

  def test_render_as_root(self):
    part = basic_parts.siblings(
        basic_parts.OnSeparateLines.build(
            [MockRenderableTreePart('A'), '', '', MockRenderableTreePart('B')]
        )
    )
    with self.subTest('text_root_default'):
      self.assertEqual(
          lowering.render_to_text_as_root(part),
          '[mock A, e:y r:n]\n[mock B, e:y r:n]',
      )
    with self.subTest('text_root_no_strip'):
      self.assertEqual(
          lowering.render_to_text_as_root(
              part,
              strip_trailing_whitespace=False,
              strip_whitespace_lines=False,
          ),
          '\n[mock A, e:y r:n]\n\n\n[mock B, e:y r:n]\n',
      )
    with self.subTest('text_root_roundtrip'):
      self.assertEqual(
          lowering.render_to_text_as_root(part, roundtrip=True),
          '[mock A, e:y r:y]\n[mock B, e:y r:y]',
      )
    with self.subTest('html_root'):
      # We don't test the full HTML source since it includes implementation
      # details; just test that it executes correctly and does contain the
      # rendered objects somewhere.
      html_output = lowering.render_to_html_as_root(part)
      self.assertContainsInOrder(
          ['[mock A, b:y]', '[mock B, b:y]'], html_output
      )

  def test_build_qualified_type_name(self):
    type_registries.update_registries_for_imports()
    part = common_structures.maybe_qualified_type_name(jax.ShapeDtypeStruct)
    self.assertEqual(
        lowering.render_to_text_as_root(part),
        'ShapeDtypeStruct',
    )
    self.assertEqual(
        lowering.render_to_text_as_root(part, roundtrip=True),
        'jax.ShapeDtypeStruct',
    )

  @parameterized.named_parameters(
      dict(
          testcase_name='build_one_line_tree_node_1',
          part=common_structures.build_one_line_tree_node(
              line='Some contents',
          ),
          expected_collapsed='Some contents',
          expected_expanded='Some contents',
      ),
      dict(
          testcase_name='build_one_line_tree_node_2',
          part=common_structures.build_one_line_tree_node(
              line=basic_parts.RenderableAndLineAnnotations(
                  MockRenderableTreePart('A'),
                  annotations=basic_parts.Text('# line comment'),
              ),
              path=(),
          ),
          # No foldable, so nothing to collapse; child is still "expanded".
          expected_collapsed='[mock A, e:y r:n]',
          expected_expanded='[mock A, e:y r:n]',
          expected_annotations='# line comment',
      ),
      dict(
          testcase_name='build_foldable_tree_node_from_children_1',
          part=common_structures.build_foldable_tree_node_from_children(
              prefix='Wrapper(',
              children=[
                  'foo',
                  MockRenderableTreePart('A'),
              ],
              suffix=')',
              comma_separated=False,
          ),
          expected_collapsed='Wrapper(foo[mock A, e:n r:n])',
          expected_expanded='Wrapper(\n  foo\n  [mock A, e:y r:n]\n)',
      ),
      dict(
          testcase_name='build_foldable_tree_node_from_children_2',
          part=common_structures.build_foldable_tree_node_from_children(
              prefix='Wrapper(',
              children=[
                  'foo',
                  MockRenderableTreePart('A'),
              ],
              suffix=')',
              path=(),
              comma_separated=True,
          ),
          expected_collapsed='Wrapper(foo, [mock A, e:n r:n])',
          expected_expanded='Wrapper(\n  foo,\n  [mock A, e:y r:n],\n)',
      ),
      dict(
          testcase_name='build_foldable_tree_node_from_children_3',
          part=common_structures.build_foldable_tree_node_from_children(
              prefix='Wrapper(',
              children=[
                  'foo',
                  basic_parts.RenderableAndLineAnnotations(
                      MockRenderableTreePart('A'),
                      annotations=basic_parts.Text('# line comment'),
                  ),
              ],
              suffix=')',
              path=(),
              comma_separated=True,
              force_trailing_comma=True,
          ),
          expected_collapsed='Wrapper(foo, [mock A, e:n r:n],)',
          expected_expanded=(
              'Wrapper(\n  foo,\n  [mock A, e:y r:n],# line comment\n)'
          ),
      ),
      dict(
          testcase_name='build_custom_foldable_tree_node',
          part=common_structures.build_custom_foldable_tree_node(
              label=basic_parts.Text('foo'),
              contents=basic_parts.siblings(
                  'bar',
                  basic_parts.OnSeparateLines.build([
                      'foo',
                      basic_parts.RenderableAndLineAnnotations(
                          MockRenderableTreePart('A'),
                          annotations=basic_parts.Text('# line comment'),
                      ),
                  ]),
              ),
              path=(),
          ),
          expected_collapsed='foobarfoo[mock A, e:n r:n]',
          expected_expanded='foobar\nfoo\n[mock A, e:y r:n]# line comment',
      ),
  )
  def test_render_common_structure(
      self,
      part: basic_parts.RenderableAndLineAnnotations,
      *,
      expected_collapsed: str,
      expected_expanded: str,
      expected_annotations: str = '',
  ):
    with self.subTest('annotations'):
      self.assertEqual(
          lowering.render_to_text_as_root(part.annotations),
          expected_annotations,
      )

    for foldable in part.renderable.foldables_in_this_part():
      foldable.set_expand_state(basic_parts.ExpandState.COLLAPSED)

    with self.subTest('text_collapsed'):
      self.assertEqual(
          lowering.render_to_text_as_root(part.renderable),
          expected_collapsed,
      )

    for foldable in part.renderable.foldables_in_this_part():
      foldable.set_expand_state(basic_parts.ExpandState.EXPANDED)

    with self.subTest('text_expanded'):
      self.assertEqual(
          lowering.render_to_text_as_root(part.renderable),
          expected_expanded,
      )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Global registries for adding treescope support to external types.

This module defines global registries which can be used to add treescope support
for new types that it does not natively support, or types defined in libraries
that may not be installed.

These registries are intended to be used by either the module that defines the
objects being registered, or by `treescope` itself. If a type already has
a global registry entry, you should generally avoid modifying it. This is
because the registries are defined as global variables, without a mechanism for
resolving conflicts between multiple entries. If you would like to customize the
rendering of a type that treescope already supports, you should generally either
define your own treescope renderer object and use it directly, or override the
default renderer or autovisualizer (`treescope.active_renderer` and
`treescope.active_autovisualizer`) using the `set_scoped` and `set_globally`
methods. This will take precedence over any global registry entries.
"""

from __future__ import annotations

import abc
import importlib
import sys
import types
from typing import Any, TypeVar

from treescope import ndarray_adapters
from treescope import renderers
from treescope._internal import object_inspection

T = TypeVar("T")


NDARRAY_ADAPTER_REGISTRY: dict[
    type[Any], ndarray_adapters.NDArrayAdapter[Any]
] = {}
"""Global registry of NDArray adapters, keyed by type.

The value for a given type should be an instance of `NDArrayAdapter`, and will
be used to render any arrays of that type.

If a type is not present in this registry, the entries of that type's `__mro__`
will also be searched. Additionally, virtual base classes will be checked if
the abstract base class is in `VIRTUAL_BASE_CLASSES`.

NDArray adapters are usually looked up using the `lookup_ndarray_adapter`
function, which will also check for the __treescope_ndarray_adapter__ method
on the type.
"""

TREESCOPE_HANDLER_REGISTRY: dict[type[Any], renderers.TreescopeNodeHandler] = {}
"""Global registry of custom treescope handlers, keyed by type.

If a type is not present in this registry, the entries of that type's `__mro__`
will also be searched. Additionally, virtual base classes will be checked if
the abstract base class is in `VIRTUAL_BASE_CLASSES`.

The handler itself will be passed the object, and can either return a treescope
rendering or the `NotImplemented` sentinel, just like an ordinary treescope
handler.

This registry is primarily intended to add treescope support to custom types
without requiring the type to be modified. If you can modify the type, you can
instead define the `__treescope_repr__` method on the type itself; this has
precedence over the registry.
"""

VIRTUAL_BASE_CLASSES: list[type[abc.ABC]] = []
"""List of abstract base classes that should be checked for virtual subclasses.

This list should contain a list of abstract base classes that have virtual
subclasses (defined using the ``.register`` method), and which appear in the
global type registries `NDARRAY_ADAPTER_REGISTRY` or
`TREESCOPE_HANDLER_REGISTRY`. If a type is a subclass of one of these base
classes, the corresponding registry entry will be used.
"""

IMMUTABLE_TYPES_REGISTRY: dict[type[Any], bool] = {
    types.FunctionType: True,
    types.MethodType: True,
    types.ModuleType: True,
    type: True,
    type(None): True,
    type(NotImplemented): True,
    type(Ellipsis): True,
}
"""Global registry of non-hashable types that are considered immutable.

By default, treescope will detect repeated values of any non-hashable type and
render a warning that they are shared across different parts of the tree. This
is intended to help catch bugs in which a value is accidentally shared between
different parts of a tree, which could cause problems when the tree is mutated.

Some types are not hashable, but are still immutable. For instance, `jax.Array`
is immutable and can be safely shared. This set is used to suppress the
"shared value" warning for these types.
"""

_LAZY_MODULE_SETUP_FUNCTIONS: dict[str, tuple[str, str]] = {
    # Note: Numpy is always imported because it is used by the core array
    # rendering system, but we define its setup function here as well for
    # consistency with the other array modules.
    "numpy": (
        "treescope.external.numpy_support",
        "set_up_treescope",
    ),
    "jax": (
        "treescope.external.jax_support",
        "set_up_treescope",
    ),
    "torch": (
        "treescope.external.torch_support",
        "set_up_treescope",
    ),
}
"""Delayed setup functions that run only once a module is imported.

This dictionary maps module name keys to a ``(setup_module, setup_attribute)``
tuple, where ``setup_module`` is the name of a module and ``setup_attribute``
is the name of a zero-argument function in that module that can be used to set
up support for the key module.

When `update_registries_for_imports` is called (usually at the start of
rendering an object), if any of the key modules have already been imported,
the corresponding setup module will be imported as well
and the setup attribute will be called. This function can then modify the
global values `NDARRAY_ADAPTER_REGISTRY`, `TREESCOPE_HANDLER_REGISTRY`,
`IMMUTABLE_TYPES_SET`, or `VIRTUAL_BASE_CLASSES` to add support for this module.
It can also register the public API of the module using
`treescope.canonical_aliases`, if applicable.

After being called, the setup function will be removed from this dictionary.
"""


def update_registries_for_imports():
  """Updates registries by running setup logic for newly-imported modules."""
  for module_name, (setup_module, setup_attribute) in tuple(
      _LAZY_MODULE_SETUP_FUNCTIONS.items()
  ):
    if module_name in sys.modules:
      module = importlib.import_module(setup_module)
      setup_fn = getattr(module, setup_attribute)
      setup_fn()
      del _LAZY_MODULE_SETUP_FUNCTIONS[module_name]


def _lookup_by_mro(
    registry: dict[type[Any], T], candidate_type: type[Any]
) -> T | None:
  """Looks up the given type in the given registry, or in its base classes.

  This function looks up the given type in the given registry, or in the
  registry for any base class of the given type, in method resolution order.

  If no concrete base class is found in the registry, each of the entries of
  `VIRTUAL_BASE_CLASSES` will be checked to see if it is a virtual base class.
  The first such base class that has an entry in the registry will be used.

  Args:
    registry: The registry to look up in.
    candidate_type: The type to look up.

  Returns:
    The value associated with the given type (or a base class of it) in the
    given registry, or None if no entry was found.
  """
  for supertype in candidate_type.__mro__:
    if supertype in registry:
      return registry[supertype]
  for base_class in VIRTUAL_BASE_CLASSES:
    if issubclass(candidate_type, base_class) and base_class in registry:
      return registry[base_class]
  return None


def lookup_immutability_for_type(candidate_type: type[Any]) -> bool:
  """Checks if an object is marked as immutable in the global registry.

  This function can be used to look up whether an object should be considered
  immutable for treescope according to `IMMUTABLE_TYPES_REGISTRY`. The status
  will be looked up by the type of the given object, any of its base classes, or
  any virtual base classes listed in `VIRTUAL_BASE_CLASSES`.

  Args:
    candidate_type: The type to check immutability of.

  Returns:
    True if this type is registered as immutable, False otherwise.
  """
  return bool(_lookup_by_mro(IMMUTABLE_TYPES_REGISTRY, candidate_type))


def lookup_treescope_handler_for_type(
    candidate_type: type[Any],
) -> renderers.TreescopeNodeHandler | None:
  """Looks up a treescope handler for the given type.

  This function can be used to look up a treescope handler for an object using
  the global registry `TREESCOPE_HANDLER_REGISTRY`. The handler will be looked
  up by the type of the given object, any of its base classes, or any virtual
  base classes listed in `VIRTUAL_BASE_CLASSES`.

  This function does NOT check for methods on the type; those should be checked
  separately.

  Args:
    candidate_type: The type to look up a handler for.

  Returns:
    A treescope handler for the given type, or None if no handler was found.
  """
  return _lookup_by_mro(TREESCOPE_HANDLER_REGISTRY, candidate_type)


def lookup_ndarray_adapter(
    possible_array: Any,
) -> ndarray_adapters.NDArrayAdapter[Any] | None:
  """Looks up an NDArray adapter for the given type.

  This function looks for an NDArray adapter by first checking for the
  `__treescope_ndarray_adapter__` method on the type, and then by looking up
  the type in the global registry `NDARRAY_ADAPTER_REGISTRY`.

  Args:
    possible_array: The object to look up an adapter for.

  Returns:
    An NDArray adapter for the given type, or None if no adapter was found.
  """
  has_adapter_method = object_inspection.safely_get_real_method(
      possible_array, "__treescope_ndarray_adapter__"
  )
  if has_adapter_method:
    return has_adapter_method()
  else:
    return _lookup_by_mro(NDARRAY_ADAPTER_REGISTRY, type(possible_array))



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Renderer objects and handler definitions."""
from __future__ import annotations

import contextlib
import dataclasses
import functools
import traceback
import typing
from typing import Any, Callable, Iterable
import warnings

from treescope import layout_algorithms
from treescope import lowering
from treescope import rendering_parts


class TreescopeSubtreeRenderer(typing.Protocol):
  """Protocol for a (recursive) subtree renderer.

  Implementations of this protocol render full subtrees to HTML tags. Handlers
  take one of these as an argument and use it to render their children.
  """

  def __call__(
      self,
      node: Any,
      path: str | None = None,
  ) -> rendering_parts.RenderableAndLineAnnotations:
    """Signature for a (recursive) subtree renderer.

    Args:
      node: The node to render.
      path: Optionally, a path to this node, represented as a string that can be
        used to reach this node from the root (e.g. ".foo.bar['baz']")

    Returns:
      A representation of the object as a renderable tree part in the
      intermediate representation.
    """


class TreescopeNodeHandler(typing.Protocol):
  """Protocol for a rendering handler for a particular node type.

  Rendering handlers run in order (from most specific to least specific),
  until one of them is able to handle the given node. Handlers should return
  NotImplemented for types of node that they are unable to handle.
  """

  def __call__(
      self,
      node: Any,
      path: str | None,
      subtree_renderer: TreescopeSubtreeRenderer,
  ) -> (
      rendering_parts.RenderableTreePart
      | rendering_parts.RenderableAndLineAnnotations
      | type(NotImplemented)
  ):
    """Signature for a rendering handler for a particular node type.

    Args:
      node: The node to render.
      path: Optionally, a path to this node, represented as a string that can be
        used to reach this node from the root (e.g. ".foo.bar['baz']"). Handlers
        should extend this path and pass it to `subtree_renderer` when rendering
        each of its children.
      subtree_renderer: A recursive renderer for subtrees of this node. This
        should be used to render the children of this node into HTML tags to
        include in the rendering of this node. Should only be called on nodes
        that are children of `node`, to avoid an infinite loop.

    Returns:
      A representation of the object as a tree part, or NotImplemented
      if this handler doesn't support rendering this object (indicating that
      we should try the next handler). Tree parts may have optional end-of-line
      annotations.
    """


class TreescopeCustomWrapperHook(typing.Protocol):
  """Protocol for a custom wrapper hook.

  Custom wrapper hooks are more powerful versions of node handlers. They run
  on every node, and are given control over how that node is rendered. In
  particular, they can modify the arguments that are passed to subsequent
  wrappers and ordinary handlers by calling ``node_renderer`` with modified
  values, and can also insert additional HTML tags around the rendered result.

  Generally, custom wrapper hooks should call ``node_renderer`` exactly once
  with the same ``node`` argument and the same ``path``.
  """

  def __call__(
      self,
      node: Any,
      path: str | None,
      node_renderer: TreescopeSubtreeRenderer,
  ) -> (
      rendering_parts.RenderableTreePart
      | rendering_parts.RenderableAndLineAnnotations
      | type(NotImplemented)
  ):
    """Signature for a custom wrapper hook.

    Args:
      node: The node that we are rendering.
      path: Optionally, a path to this node as a string; see
        `TreescopeNodeHandler`.
      node_renderer: The inner renderer for this node. This can be used to
        render `node` itself according to the ordinary logic. Usually the
        wrapper hook should call `node_renderer` once with `node`, possibly
        customizing its arguments, and then postprocess the return value.

    Returns:
      A modified rendering of the object, or NotImplemented if this
      hook does not want to modify this rendering at all. (Returning
      NotImplemented)
    """


@dataclasses.dataclass
class TreescopeRenderer:
  """A renderer object with a configurable set of handlers.

  Attributes:
    handlers: List of handlers, ordered from most specific to least specific.
      Handlers will be tried in order until one returns something other than
      NotImplemented.
    wrapper_hooks: List of wrapper hooks, ordered from outermost to innermost.
      All wrapper hooks will run in order on every node.
    context_builders: List of functions that build rendering context managers.
      Each of these context managers will be entered before rendering the tree,
      and will be exited after converting the returned tags to an HTML string.
      The main use for this is to set up shared state needed by the handlers or
      wrapper hooks (e.g. using a `treescope.context.ContextualValue`). Context
      builders can also return None instead of a context manager, if they just
      need to do processing at the start of rendering and don't need to clean
      up.
  """

  handlers: list[TreescopeNodeHandler]
  wrapper_hooks: list[TreescopeCustomWrapperHook]
  context_builders: list[
      Callable[[], contextlib.AbstractContextManager[Any] | None]
  ]

  def extended_with(
      self,
      *,
      handlers: Iterable[TreescopeNodeHandler] = (),
      wrapper_hooks: Iterable[TreescopeCustomWrapperHook] = (),
      context_builders: Iterable[
          Callable[[], contextlib.AbstractContextManager[Any] | None]
      ] = (),
  ) -> TreescopeRenderer:
    """Extends a renderer with additional steps, returning a new renderer.

    Args:
      handlers: New handlers to insert. These will be inserted at the beginning
        of the handler list, to allow them to override existing handlers.
      wrapper_hooks: New wrapper hooks. These will be inserted before the
        existing wrapper hooks, to allow them to override existing hooks.
      context_builders: New context builders. These will be inserted after the
        existing context builders.

    Returns:
      A copy of this renderer extended with the new parts.
    """
    return TreescopeRenderer(
        handlers=list(handlers) + self.handlers,
        wrapper_hooks=list(wrapper_hooks) + self.wrapper_hooks,
        context_builders=self.context_builders + list(context_builders),
    )

  def _render_subtree(
      self,
      ignore_exceptions: bool,
      rendering_stack: list[Any],
      already_executed_wrapper_count: int,
      node: Any,
      path: str | None = None,
  ) -> rendering_parts.RenderableAndLineAnnotations:
    """Renders a specific subtree using the renderer.

    Args:
      ignore_exceptions: Whether to catch and ignore errors in handlers or
        postprocessors.
      rendering_stack: A list of the ancestors of this node, used to check for
        and break cyclic references.
      already_executed_wrapper_count: The number of wrappers we have already
        called for this node.
      node: See `TreescopeSubtreeRenderer.__call__`.
      path: See `TreescopeSubtreeRenderer.__call__`.

    Returns:
      A renderable representation of the object.
    """
    while already_executed_wrapper_count < len(self.wrapper_hooks):
      # We have hooks we haven't processed. Defer to the hook.
      # We pass a reference to this method back to the wrapper to allow it to
      # resume execution at the next wrapper.
      hook = self.wrapper_hooks[already_executed_wrapper_count]
      render_without_this_hook = functools.partial(
          self._render_subtree,
          ignore_exceptions,
          rendering_stack,
          already_executed_wrapper_count + 1,
      )
      try:
        postprocessed_result = hook(
            node=node, path=path, node_renderer=render_without_this_hook
        )
        if postprocessed_result is NotImplemented:
          # Skip this hook and defer to the next one (or to a handler)
          already_executed_wrapper_count += 1
          continue

        elif isinstance(
            postprocessed_result, rendering_parts.RenderableAndLineAnnotations
        ):
          return postprocessed_result
        elif isinstance(
            postprocessed_result, rendering_parts.RenderableTreePart
        ):
          return rendering_parts.RenderableAndLineAnnotations(
              renderable=postprocessed_result,
              annotations=rendering_parts.empty_part(),
          )
        else:
          raise ValueError(
              f"Wrapper hook {hook} returned a value of an unexpected type:"
              f" {postprocessed_result}"
          )
      except Exception:  # pylint: disable=broad-exception-caught
        if ignore_exceptions:
          warnings.warn(
              f"Ignoring error inside wrapper hook {hook}"
              f":\n{traceback.format_exc(limit=10)}"
          )
          already_executed_wrapper_count += 1
          continue
        else:
          raise

    # If we get here, all of the wrappers have already processed this object.

    # Use object identity to track cyclic objects.
    node_id = id(node)

    if node_id in rendering_stack:
      # Cycle! This object contains itself.
      return rendering_parts.RenderableAndLineAnnotations(
          renderable=rendering_parts.error_color(
              rendering_parts.text(
                  f"<cyclic reference to {type(node).__name__} at"
                  f" 0x{node_id:x}>"
              )
          ),
          annotations=rendering_parts.empty_part(),
      )
    else:
      # Track cyclic references. We use `try: ... finally: ...` to ensure we
      # clean up after ourselves even if we return a value early. (There's also
      # separate internal try/catch logic for error handling.)
      rendering_stack.append(node_id)
      try:
        # Give each of our handlers a chance to render this node.
        # We pass this method as the renderer for any children of the node, and
        # reenable all of the wrapper hooks.
        rec = functools.partial(
            self._render_subtree, ignore_exceptions, rendering_stack, 0
        )
        for handler in self.handlers:
          try:
            maybe_result = handler(node=node, path=path, subtree_renderer=rec)

            if maybe_result is NotImplemented:
              # Try the next handler.
              continue
            elif isinstance(
                maybe_result, rendering_parts.RenderableAndLineAnnotations
            ):
              # Found a result!
              return maybe_result
            elif isinstance(maybe_result, rendering_parts.RenderableTreePart):
              # Wrap it with empty annotations.
              return rendering_parts.RenderableAndLineAnnotations(
                  renderable=maybe_result,
                  annotations=rendering_parts.empty_part(),
              )
            else:
              raise ValueError(
                  f"Handler {handler} returned a value of an unexpected type:"
                  f" {maybe_result}"
              )

          except Exception:  # pylint: disable=broad-exception-caught
            if ignore_exceptions:
              warnings.warn(
                  "Ignoring error while formatting value of type"
                  f" {type(node)} with"
                  f" {handler}:\n{traceback.format_exc(limit=10)}"
              )
              continue
            else:
              raise

        # If we get here, it means no handler was found.
        if ignore_exceptions:
          warnings.warn(
              f"No handler registered for a node of type {type(node)}."
          )
          # Fall back to a basic `repr` so that we still render something even
          # without a handler for it. We use the object repr because a custom
          # repr may still raise an exception if the object is in an invalid
          # state.
          return rendering_parts.RenderableAndLineAnnotations(
              renderable=rendering_parts.error_color(
                  rendering_parts.text(object.__repr__(node))
              ),
              annotations=rendering_parts.comment_color(
                  rendering_parts.text(
                      "  # Error occurred while formatting this object."
                  )
              ),
          )
        else:
          raise ValueError(
              f"No handler registered for a node of type {type(node)}."
          )

      finally:
        # Remove it from the rendering stack.
        popped_val = rendering_stack.pop()
        assert popped_val == node_id

  def to_foldable_representation(
      self,
      value: Any,
      ignore_exceptions: bool = False,
      root_keypath: str | None = "",
  ) -> rendering_parts.RenderableAndLineAnnotations:
    """Renders an object to the foldable intermediate representation.

    Args:
      value: Value to render.
      ignore_exceptions: Whether to catch errors during rendering of subtrees
        and show a fallback for those subtrees, instead of failing the entire
        renderer. Best used in contexts where `to_foldable_representation` is
        not being called directly by the user, e.g. when registering this as a
        default pretty-printer.
      root_keypath: Keypath to the value being rendered. Can be None to disable
        all keypaths during rendering.

    Returns:
      Renderable representation of the object.
    """
    with contextlib.ExitStack() as stack:
      # Set up any scoped state contexts that handlers need.
      for context_builder in self.context_builders:
        new_context = context_builder()
        if new_context is not None:
          stack.enter_context(new_context)

      # Build the foldable HTML tag recursively.
      return self._render_subtree(
          ignore_exceptions=ignore_exceptions,
          rendering_stack=[],
          already_executed_wrapper_count=0,
          node=value,
          path=root_keypath,
      )

  def to_text(self, value: Any, roundtrip_mode: bool = False) -> str:
    """Convenience method to render an object to text.

    Args:
      value: Value to render.
      roundtrip_mode: Whether to render in roundtrip mode.

    Returns:
      A text representation of the object.
    """
    foldable_ir = rendering_parts.build_full_line_with_annotations(
        self.to_foldable_representation(value)
    )
    layout_algorithms.expand_for_balanced_layout(foldable_ir)
    return lowering.render_to_text_as_root(foldable_ir, roundtrip_mode)

  def to_html(self, value: Any, roundtrip_mode: bool = False) -> str:
    """Convenience method to render an object to HTML.

    Args:
      value: Value to render.
      roundtrip_mode: Whether to render in roundtrip mode.

    Returns:
      HTML source code for the foldable representation of the object.
    """
    foldable_ir = rendering_parts.build_full_line_with_annotations(
        self.to_foldable_representation(value)
    )
    layout_algorithms.expand_for_balanced_layout(foldable_ir)
    return lowering.render_to_html_as_root(foldable_ir, roundtrip_mode)



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration file for the Sphinx documentation builder."""

# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

# pylint: disable=g-bad-import-order
# pylint: disable=g-import-not-at-top
import inspect
import logging
import os
import subprocess
import sys
import typing

import treescope

from sphinx.util import logging as sphinx_logging

typing.get_type_hints = lambda obj, *unused: obj.__annotations__
sys.path.insert(0, os.path.abspath('../'))
sys.path.append(os.path.abspath('ext'))

from sphinxcontrib import katex


class IgnoreBangWarningFilter(logging.Filter):

  def filter(self, record):
    msg = record.getMessage()
    if msg.startswith('Lexing literal_block') and msg.endswith(
        'as "python" resulted in an error at token: \'!\'. Retrying in relaxed'
        ' mode.'
    ):
      return False
    else:
      return True


sphinx_logging.getLogger('sphinx.highlighting').logger.addFilter(
    IgnoreBangWarningFilter()
)

try:
  # Look up the current git hash so we can link to it.
  git_hash_bytes = subprocess.check_output(['git', 'rev-parse', 'HEAD'])
  source_git_hash = git_hash_bytes.decode().strip()
except Exception:  # pylint: disable=broad-exception-caught
  source_git_hash = None

# -- Project information -----------------------------------------------------

project = 'Treescope'
copyright = '2024, The Treescope Authors'  # pylint: disable=redefined-builtin
author = 'the Treescope Authors'

# -- General configuration ---------------------------------------------------

master_doc = 'index'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.intersphinx',
    'sphinx.ext.linkcode',
    'sphinx.ext.napoleon',
    'sphinxcontrib.katex',
    'myst_nb',
    'sphinx_contributors',
    'treescope_sphinx_customizations',
    'nb_output_cell_to_iframe',
    'hoverxref.extension',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

default_role = 'py:obj'

# -- Options for napoleon -----------------------------------------------------

napoleon_google_docstring = True
napoleon_use_rtype = True
napoleon_attr_annotations = True
napoleon_use_ivar = True

# typehints_use_signature = True
# typehints_use_signature_return = True

# -- Options for autodoc / autosummary -----------------------------------------

autoclass_content = 'class'
autodoc_class_signature = 'separated'
autodoc_default_options = {
    'exclude-members': (
        '__repr__, __str__, __weakref__, __hash__, __delattr__, __eq__,'
        ' __setattr__, __subclasshook__'
    ),
}
autodoc_inherit_docstrings = False

autosummary_ignore_module_all = False
autosummary_generate = ['api/treescope.rst', '_autogen_root.rst']

# -- Options for HTML output -------------------------------------------------

html_title = 'treescope'

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_book_theme'

html_theme_options = {
    'repository_url': 'https://github.com/google-deepmind/treescope',
    'use_repository_button': True,
    'use_issues_button': True,
    'show_toc_level': 4,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_css_files = ['custom.css']
html_js_files = ['custom.js']
# html_favicon = '_static/favicon.ico'

# -- Options for myst -------------------------------------------------------
myst_enable_extensions = [
    'amsmath',
    'dollarmath',
]
# nb_execution_mode can be overridden on a notebook level via .ipynb metadata
nb_execution_mode = 'auto'
nb_execution_allow_errors = False
# nb_execution_excludepatterns = ['notebooks/*']

# -- Options for katex ------------------------------------------------------

# See: https://sphinxcontrib-katex.readthedocs.io/en/0.4.1/macros.html
latex_macros = r"""
    \def \d              #1{\operatorname{#1}}
"""

# Translate LaTeX macros to KaTeX and add to options for HTML builder
katex_macros = katex.latex_defs_to_katex_macros(latex_macros)
katex_options = 'macros: {' + katex_macros + '}'

# Add LaTeX macros for LATEX builder
latex_elements = {'preamble': latex_macros}

# -- Hover cross-references ----------------------------------------------------

hoverxref_domains = [
    'py',
]
hoverxref_role_types = {
    'class': 'tooltip',
    'exc': 'tooltip',
    'func': 'tooltip',
    'mod': 'tooltip',
    'obj': 'tooltip',
}

# -- Source code links -------------------------------------------------------


def linkcode_resolve(domain, info):
  """Resolve a GitHub URL corresponding to Python object."""
  if domain != 'py':
    return None

  try:
    mod = sys.modules[info['module']]
  except ImportError:
    return None

  obj = mod
  try:
    for attr in info['fullname'].split('.'):
      obj = getattr(obj, attr)
  except AttributeError:
    return None
  else:
    obj = inspect.unwrap(obj)

  try:
    filename = inspect.getsourcefile(obj)
  except TypeError:
    return None

  try:
    source, lineno = inspect.getsourcelines(obj)
  except OSError:
    return None

  if source_git_hash is not None:
    git_identifier = source_git_hash
  else:
    git_identifier = 'main'
  relpath = os.path.relpath(filename, start=os.path.dirname(treescope.__file__))
  return (
      f'https://github.com/google-deepmind/treescope/blob/{git_identifier}/treescope/'
      f'{relpath}#L{lineno}#L{lineno + len(source) - 1}'
  )


# -- Intersphinx configuration -----------------------------------------------

intersphinx_mapping = {
    'jax': ('https://jax.readthedocs.io/en/latest/', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'penzai': ('https://penzai.readthedocs.io/en/stable/', None),
    'python': ('https://docs.python.org/3/', None),
}

source_suffix = ['.rst', '.md', '.ipynb']



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Lazy setup logic for adding PyTorch support to treescope."""

from __future__ import annotations

import keyword
import typing

import numpy as np
from treescope import context
from treescope import formatting_util
from treescope import lowering
from treescope import ndarray_adapters
from treescope import renderers
from treescope import rendering_parts
from treescope import type_registries

# pylint: disable=import-outside-toplevel
try:
  import torch  # pytype: disable=import-error
except ImportError:
  assert not typing.TYPE_CHECKING
  torch = None
# pylint: enable=import-outside-toplevel

show_dynamic_attributes: context.ContextualValue[bool] = (
    context.ContextualValue(
        module=__name__,
        qualname="show_dynamic_attributes",
        initial_value=True,
    )
)
"""Whether to inspect and show all non-private attributes of Torch modules.

If set to True, when rendering a Torch module, we will walk all of its
attributes (the entries in its `__dict__`) and render every attribute that does
not start with an underscore. If set to False, we will defer to the `extra_repr`
method instead, which matches default Torch repr behavior.
"""


def _tensor_to_numpy(tensor: torch.Tensor) -> np.ndarray:
  """Converts a tensor to a numpy array, safely handling 16-bit float dtypes."""
  if tensor.dtype == torch.bfloat16 or tensor.dtype == torch.float16:
    return tensor.float().numpy(force=True)
  return tensor.numpy(force=True)


def _truncate_and_copy(
    array_source: torch.Tensor,
    array_dest: np.ndarray,
    prefix_slices: tuple[slice, ...],
    remaining_edge_items_per_axis: tuple[int | None, ...],
) -> None:
  """Recursively copy values on the edges of a torch tensor into a numpy array.

  This function mutates the destination array in place, copying parts of input
  array into them, so that it contains a truncated versions of the original
  array.

  Args:
    array_source: Source array, which we will truncate.
    array_dest: Destination array, whose axis sizes will be either the same as
      `array_source` or of size `2 * edge_items + 1` depending on the
      truncation.
    prefix_slices: Prefix of slices for the source and destination.
    remaining_edge_items_per_axis: Number of edge items to keep for each axis,
      ignoring any axes whose slices are already computed in `source_slices`.
  """
  assert torch is not None, "PyTorch is not available."
  if not remaining_edge_items_per_axis:
    # Perform the base case slice.
    assert (
        len(prefix_slices) == len(array_source.shape) == len(array_dest.shape)
    )
    array_dest[prefix_slices] = _tensor_to_numpy(array_source[prefix_slices])
  else:
    # Recursive step.
    axis = len(prefix_slices)
    edge_items = remaining_edge_items_per_axis[0]
    if edge_items is None:
      # Don't need to slice.
      _truncate_and_copy(
          array_source=array_source,
          array_dest=array_dest,
          prefix_slices=prefix_slices + (slice(None),),
          remaining_edge_items_per_axis=remaining_edge_items_per_axis[1:],
      )
    else:
      assert array_source.shape[axis] > 2 * edge_items
      _truncate_and_copy(
          array_source=array_source,
          array_dest=array_dest,
          prefix_slices=prefix_slices + (slice(None, edge_items),),
          remaining_edge_items_per_axis=remaining_edge_items_per_axis[1:],
      )
      _truncate_and_copy(
          array_source=array_source,
          array_dest=array_dest,
          prefix_slices=prefix_slices + (slice(-edge_items, None),),
          remaining_edge_items_per_axis=remaining_edge_items_per_axis[1:],
      )


class TorchTensorAdapter(ndarray_adapters.NDArrayAdapter[torch.Tensor]):
  """NDArray adapter for Torch tensors."""

  def get_axis_info_for_array_data(
      self, array: torch.Tensor
  ) -> tuple[ndarray_adapters.AxisInfo, ...]:
    infos = []
    for i, (size, name) in enumerate(zip(array.shape, array.names)):
      if name is None:
        infos.append(ndarray_adapters.PositionalAxisInfo(i, size))
      else:
        infos.append(
            ndarray_adapters.NamedPositionalAxisInfo(
                axis_logical_index=i, axis_name=name, size=size
            )
        )
    return tuple(infos)

  def get_array_data_with_truncation(
      self,
      array: torch.Tensor,
      mask: torch.Tensor | None,
      edge_items_per_axis: tuple[int | None, ...],
  ) -> tuple[np.ndarray, np.ndarray]:
    assert torch is not None, "PyTorch is not available."
    array = array.detach()

    if mask is None:
      mask = np.ones((1,) * array.ndim, dtype=bool)

    mask = torch.as_tensor(mask).detach()

    # Broadcast mask. (Note: Broadcasting a Numpy array does not copy data.)
    mask = torch.broadcast_to(mask, array.shape)

    if edge_items_per_axis == (None,) * array.ndim:
      # No truncation.
      return _tensor_to_numpy(array), _tensor_to_numpy(mask)

    dest_shape = [
        size if edge_items is None else 2 * edge_items + 1
        for size, edge_items in zip(array.shape, edge_items_per_axis)
    ]
    array_dest = np.zeros(dest_shape, dtype=self.get_numpy_dtype(array))
    mask_dest = np.zeros(dest_shape, dtype=np.bool_)
    _truncate_and_copy(
        array_source=array,
        array_dest=array_dest,
        prefix_slices=(),
        remaining_edge_items_per_axis=edge_items_per_axis,
    )
    _truncate_and_copy(
        array_source=mask,
        array_dest=mask_dest,
        prefix_slices=(),
        remaining_edge_items_per_axis=edge_items_per_axis,
    )
    return array_dest, mask_dest

  def get_array_summary(self, array: torch.Tensor, fast: bool) -> str:
    assert torch is not None, "PyTorch is not available."
    ty = type(array)
    array = array.detach()
    typename = f"{ty.__module__}.{ty.__name__}"
    if typename == "torch.nn.parameter.Parameter":
      typename = "torch.nn.Parameter"
    output_parts = [f"{typename} "]

    output_parts.append(repr(array.dtype).removeprefix("torch."))
    name_parts = []
    for size, name in zip(array.shape, array.names):
      if name:
        name_parts.append(f"{name}:{size}")
      else:
        name_parts.append(f"{size}")
    if len(name_parts) == 1:
      output_parts.append("(" + name_parts[0] + ",)")
    else:
      output_parts.append("(" + ", ".join(name_parts) + ")")

    # Drop axis names.
    if any(name is not None for name in array.names):
      array = array.rename(None)
    size = np.prod(array.shape)
    if size > 0 and size < 100_000 and not fast:
      is_floating = array.dtype.is_floating_point
      is_bool = array.dtype == torch.bool
      is_integer = (
          not is_floating and not is_bool and not array.dtype.is_complex
      )

      if is_floating:
        isfinite = torch.isfinite(array)
        any_finite = torch.any(isfinite)
        inf_to_nan = torch.where(isfinite, array, torch.nan)
        mean = torch.nanmean(inf_to_nan)
        std = torch.nanmean(torch.square(inf_to_nan - mean))

        if any_finite:
          output_parts.append(f" ≈{float(mean):.2} ±{float(std):.2}")
          nanmin = torch.amin(torch.where(isfinite, array, torch.inf))
          nanmax = torch.amax(torch.where(isfinite, array, -torch.inf))
          output_parts.append(f" [≥{float(nanmin):.2}, ≤{float(nanmax):.2}]")

      if is_integer:
        output_parts.append(
            f" [≥{torch.amin(array):_d}, ≤{torch.amax(array):_d}]"
        )

      if is_floating or is_integer:
        ct_zero = torch.count_nonzero(array == 0)
        if ct_zero:
          output_parts.append(f" zero:{ct_zero:_d}")

        ct_nonzero = torch.count_nonzero(array)
        if ct_nonzero:
          output_parts.append(f" nonzero:{ct_nonzero:_d}")

      if is_floating:
        ct_nan = torch.count_nonzero(torch.isnan(array))
        if ct_nan:
          output_parts.append(f" nan:{ct_nan:_d}")

        ct_inf = torch.count_nonzero(array == torch.inf)
        if ct_inf:
          output_parts.append(f" inf:{ct_inf:_d}")

        ct_neginf = torch.count_nonzero(array == -torch.inf)
        if ct_neginf:
          output_parts.append(f" -inf:{ct_neginf:_d}")

      if is_bool:
        ct_true = torch.count_nonzero(array)
        if ct_true:
          output_parts.append(f" true:{ct_true:_d}")

        ct_false = torch.count_nonzero(torch.logical_not(array))
        if ct_false:
          output_parts.append(f" false:{ct_false:_d}")

    return "".join(output_parts)

  def get_sharding_info_for_array_data(
      self, array: torch.Tensor
  ) -> ndarray_adapters.ShardingInfo | None:
    assert torch is not None, "PyTorch is not available."
    # Torch tensors are always on a single device.
    return ndarray_adapters.ShardingInfo(
        shard_shape=array.shape,
        device_index_to_shard_slices={
            array.get_device(): (slice(None),) * array.ndim
        },
        device_type=array.device.type,
    )

  def get_numpy_dtype(self, array: torch.Tensor) -> np.dtype:
    assert torch is not None, "PyTorch is not available."
    # Convert a zero-sized tensor to a numpy array to get its dtype.
    return _tensor_to_numpy(torch.zeros((0,), dtype=array.dtype)).dtype

  def should_autovisualize(self, array: torch.Tensor) -> bool:
    assert torch is not None, "PyTorch is not available."
    return array.device.type != "meta"


def render_torch_tensors(
    node: torch.Tensor,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Renders a numpy array."""
  assert torch is not None, "PyTorch is not available."
  del subtree_renderer
  assert isinstance(node, torch.Tensor)
  if node.device.type == "meta":
    # Don't render tensors on the meta device (they have no data).
    return NotImplemented
  adapter = TorchTensorAdapter()

  def _placeholder() -> rendering_parts.RenderableTreePart:
    return rendering_parts.deferred_placeholder_style(
        rendering_parts.text(adapter.get_array_summary(node, fast=True))
    )

  def _thunk(placeholder_expand_state: rendering_parts.ExpandState | None):
    # Is this array simple enough to render without a summary?
    node_repr = repr(node)
    if "\n" not in node_repr and "..." not in node_repr:
      if node_repr.startswith("tensor("):
        # Add module path, for consistency with other Treescope renderings.
        node_repr = f"torch.{node_repr}"
      rendering = rendering_parts.text(node_repr)
    else:
      if node_repr.count("\n") <= 15:
        if placeholder_expand_state is None:
          default_expand_state = rendering_parts.ExpandState.WEAKLY_EXPANDED
        else:
          default_expand_state = placeholder_expand_state
      else:
        # Always start big NDArrays in collapsed mode to hide irrelevant detail.
        default_expand_state = rendering_parts.ExpandState.COLLAPSED

      # Render it with a summary.
      summarized = adapter.get_array_summary(node, fast=False)
      rendering = rendering_parts.build_custom_foldable_tree_node(
          label=rendering_parts.abbreviation_color(
              rendering_parts.comment_color_when_expanded(
                  rendering_parts.siblings(
                      rendering_parts.fold_condition(
                          expanded=rendering_parts.text("# "),
                          collapsed=rendering_parts.text("<"),
                      ),
                      summarized,
                      rendering_parts.fold_condition(
                          collapsed=rendering_parts.text(">")
                      ),
                  )
              )
          ),
          contents=rendering_parts.fold_condition(
              expanded=rendering_parts.indented_children(
                  [rendering_parts.text(node_repr)]
              )
          ),
          path=path,
          expand_state=default_expand_state,
      ).renderable

    return rendering

  return rendering_parts.RenderableAndLineAnnotations(
      renderable=lowering.maybe_defer_rendering(
          main_thunk=_thunk,
          placeholder_thunk=_placeholder,
          expanded_newlines_for_layout=8,
      ),
      annotations=rendering_parts.build_copy_button(path),
  )


def render_torch_modules(
    node: torch.nn.Module,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Renders a torch module."""
  assert torch is not None, "PyTorch is not available."
  assert isinstance(node, torch.nn.Module)
  node_type = type(node)
  constructor = rendering_parts.siblings(
      rendering_parts.roundtrip_condition(roundtrip=rendering_parts.text("<")),
      rendering_parts.maybe_qualified_type_name(node_type),
      "(",
  )
  closing_suffix = rendering_parts.siblings(
      ")",
      rendering_parts.roundtrip_condition(roundtrip=rendering_parts.text(">")),
  )

  if hasattr(node, "__treescope_color__") and callable(
      node.__treescope_color__
  ):
    background_color, background_pattern = (
        formatting_util.parse_simple_color_and_pattern_spec(
            node.__treescope_color__(), node_type.__name__
        )
    )
  elif type(node) is torch.nn.Sequential:  # pylint: disable=unidiomatic-typecheck
    background_color, background_pattern = (
        formatting_util.parse_simple_color_and_pattern_spec(
            ("#cdcdcd", "color-mix(in oklab, #cdcdcd 25%, white)")
        )
    )
  elif type(node).forward is torch.nn.Module.forward:
    # No implementation of forward. Don't color-code; this is probably a
    # container like ModuleList or ModuleDict.
    background_color = None
    background_pattern = None
  else:
    type_string = node_type.__module__ + "." + node_type.__qualname__
    background_color = formatting_util.color_from_string(type_string)
    background_pattern = None

  children = []
  prefers_expand = False
  attr_children = None
  has_attr_children_expander = False

  # Render constant attributes.
  if show_dynamic_attributes.get():
    attr_children = []
    key_order = [
        key
        for key in vars(node)
        if not key.startswith("_") and key != "training"
    ]
    if "training" in vars(node):
      key_order.append("training")
    for attr in key_order:
      value = vars(node)[attr]
      child_path = None if path is None else f"{path}.{attr}"
      attr_children.append(
          rendering_parts.build_full_line_with_annotations(
              rendering_parts.siblings_with_annotations(
                  f"{attr}=",
                  subtree_renderer(value, path=child_path),
                  ",",
                  rendering_parts.fold_condition(
                      collapsed=rendering_parts.text(" ")
                  ),
              )
          )
      )
    if len(attr_children) <= 1:
      children.extend(attr_children)
    else:
      has_attr_children_expander = True
      children.append(
          rendering_parts.build_custom_foldable_tree_node(
              label=rendering_parts.fold_condition(
                  expanded=rendering_parts.comment_color(
                      rendering_parts.text("# Attributes:")
                  ),
              ),
              contents=rendering_parts.on_separate_lines(attr_children),
              path=None,
              expand_state=rendering_parts.ExpandState.COLLAPSED,
          )
      )
  else:
    extra_repr = node.extra_repr()
    if extra_repr:
      if not extra_repr.strip().endswith(","):
        extra_repr = extra_repr + ", "
      if "\n" in extra_repr:
        children.append(
            rendering_parts.on_separate_lines(extra_repr.split("\n"))
        )
        prefers_expand = True
      else:
        children.append(rendering_parts.text(extra_repr))

  # Render parameters and buffers
  for group_name, group in (
      ("Parameters", node.named_parameters(recurse=False)),
      ("Buffers", node.named_buffers(recurse=False)),
  ):
    group = list(group)
    if group:
      children.append(
          rendering_parts.fold_condition(
              expanded=rendering_parts.comment_color(
                  rendering_parts.text(f"# {group_name}:")
              )
          )
      )
      for name, value in group:
        child_path = None if path is None else f"{path}.{name}"
        children.append(
            rendering_parts.build_full_line_with_annotations(
                rendering_parts.siblings_with_annotations(
                    f"{name}=",
                    subtree_renderer(value, path=child_path),
                    ",",
                    rendering_parts.fold_condition(
                        collapsed=rendering_parts.text(" ")
                    ),
                )
            )
        )

  # Render submodules.
  submodules = list(node.named_children())
  if submodules:
    children.append(
        rendering_parts.fold_condition(
            expanded=rendering_parts.comment_color(
                rendering_parts.text("# Child modules:")
            )
        )
    )
    for name, submod in submodules:
      prefers_expand = True
      if name.isidentifier() and not keyword.iskeyword(name):
        child_path = None if path is None else f"{path}.{name}"
        keystr = f"{name}="
      else:
        child_path = f"{path}.get_submodule({repr(name)})"
        keystr = f"({name}): "
      children.append(
          rendering_parts.build_full_line_with_annotations(
              rendering_parts.siblings_with_annotations(
                  keystr,
                  subtree_renderer(submod, path=child_path),
                  ",",
                  rendering_parts.fold_condition(
                      collapsed=rendering_parts.text(" ")
                  ),
              )
          )
      )

  # If there are only dynamic attributes, don't add the level of indirection.
  if has_attr_children_expander and len(children) == 1:
    children = attr_children

  # Heuristic: If a module doesn't have any submodules, mark it collapsed, to
  # match the behavior of PyTorch repr.
  if prefers_expand:
    expand_state = rendering_parts.ExpandState.WEAKLY_EXPANDED
  else:
    expand_state = rendering_parts.ExpandState.COLLAPSED

  return rendering_parts.build_foldable_tree_node_from_children(
      prefix=constructor,
      children=children,
      suffix=closing_suffix,
      path=path,
      background_color=background_color,
      background_pattern=background_pattern,
      expand_state=expand_state,
  )


def set_up_treescope():
  """Sets up treescope to render PyTorch objects."""
  if torch is None:
    raise RuntimeError(
        "Cannot set up PyTorch support in treescope: PyTorch cannot be"
        " imported."
    )
  type_registries.NDARRAY_ADAPTER_REGISTRY[torch.Tensor] = TorchTensorAdapter()
  type_registries.TREESCOPE_HANDLER_REGISTRY[torch.Tensor] = (
      render_torch_tensors
  )
  type_registries.TREESCOPE_HANDLER_REGISTRY[torch.nn.Module] = (
      render_torch_modules
  )



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for treescope's canonical_aliases."""

import types

from absl.testing import absltest
from absl.testing import parameterized
import jax
import numpy
from treescope import canonical_aliases
from treescope import type_registries
from tests import fixtures as fixture_parent
from tests.fixtures import treescope_examples_fixture as fixture_lib


def fresh_canonical_aliases():
  return canonical_aliases._alias_environment.set_scoped(
      canonical_aliases.CanonicalAliasEnvironment({})
  )


class TreescopeCanonicalAliasesTest(parameterized.TestCase):

  def test_module_attribute_path_string(self):
    self.assertEqual(
        str(canonical_aliases.ModuleAttributePath("foo.bar", ("baz", "Qux"))),
        "foo.bar.baz.Qux",
    )
    self.assertEqual(
        str(canonical_aliases.ModuleAttributePath("__main__", ("Foo",))),
        "Foo",
    )

  def test_module_attribute_path_retrieve(self):
    path = canonical_aliases.ModuleAttributePath(
        fixture_lib.__name__, ("DataclassWithTwoChildren",)
    )
    self.assertIs(path.retrieve(), fixture_lib.DataclassWithTwoChildren)

  def test_module_attribute_path_retrieve_nested(self):
    path = canonical_aliases.ModuleAttributePath(
        fixture_lib.__name__, ("SomeOuterClass", "NestedClass")
    )
    self.assertIs(path.retrieve(), fixture_lib.SomeOuterClass.NestedClass)

  def test_module_attribute_path_retrieve_error(self):
    path = canonical_aliases.ModuleAttributePath(
        fixture_lib.__name__, "NonexistentClass"
    )
    with self.assertRaises(ValueError):
      path.retrieve()

    self.assertIsNone(path.retrieve(forgiving=True))

  def test_local_name_path_str(self):
    self.assertEqual(
        str(canonical_aliases.LocalNamePath("foo", ("baz", "Qux"))),
        "foo.baz.Qux",
    )
    self.assertEqual(str(canonical_aliases.LocalNamePath("bar", ())), "bar")

  def test_local_name_path_retrieve(self):
    path = canonical_aliases.LocalNamePath("foo", ("bar", "baz"))
    self.assertEqual(
        path.retrieve(
            {"foo": types.SimpleNamespace(bar=types.SimpleNamespace(baz=1234))}
        ),
        1234,
    )

  def test_add_and_look_up_alias(self):
    with fresh_canonical_aliases():
      self.assertIsNone(
          canonical_aliases.lookup_alias(
              fixture_lib.SomeOuterClass.NestedClass,
              infer_from_attributes=False,
          ),
      )
      canonical_aliases.add_alias(
          fixture_lib.SomeOuterClass.NestedClass,
          canonical_aliases.ModuleAttributePath(
              fixture_lib.__name__,
              ("SomeOuterClass", "NestedClass"),
          ),
      )
      self.assertEqual(
          canonical_aliases.lookup_alias(
              fixture_lib.SomeOuterClass.NestedClass,
              infer_from_attributes=False,
          ),
          canonical_aliases.ModuleAttributePath(
              fixture_lib.__name__,
              ("SomeOuterClass", "NestedClass"),
          ),
      )

  def test_infer_alias_from_attributes(self):
    with fresh_canonical_aliases():
      self.assertEqual(
          canonical_aliases.lookup_alias(
              fixture_lib.SomeOuterClass.NestedClass,
              infer_from_attributes=True,
          ),
          canonical_aliases.ModuleAttributePath(
              fixture_lib.__name__,
              ("SomeOuterClass", "NestedClass"),
          ),
      )

  def test_add_uninferabble_class(self):
    # ClassDefinedInAFunction is defined inside a function, so it's __qualname__
    # includes "<locals>"
    with fresh_canonical_aliases():
      self.assertIsNone(
          canonical_aliases.lookup_alias(
              fixture_lib.ClassDefinedInAFunction,
              infer_from_attributes=True,
          ),
      )
      canonical_aliases.add_alias(
          fixture_lib.ClassDefinedInAFunction,
          canonical_aliases.ModuleAttributePath(
              fixture_lib.__name__,
              ("ClassDefinedInAFunction",),
          ),
      )
      self.assertEqual(
          canonical_aliases.lookup_alias(
              fixture_lib.ClassDefinedInAFunction,
              infer_from_attributes=True,
          ),
          canonical_aliases.ModuleAttributePath(
              fixture_lib.__name__,
              ("ClassDefinedInAFunction",),
          ),
      )

  def test_add_bad_alias(self):
    with fresh_canonical_aliases():
      with self.assertRaises(ValueError):
        canonical_aliases.add_alias(
            fixture_lib.SomeOuterClass.NestedClass,
            canonical_aliases.ModuleAttributePath(
                fixture_lib.__name__,
                ("not_a_correct_path", "NestedClass"),
            ),
        )

  def test_outdated_alias(self):
    # sanity checks
    self.assertEqual(
        fixture_lib.OriginalRedefinedClass.__name__,
        "RedefinedClass",
    )
    self.assertIsNot(
        fixture_lib.RedefinedClass,
        fixture_lib.OriginalRedefinedClass,
    )
    # Make sure canonical aliases don't get confused. (This can happen in some
    # cases due to module reloads, for instance.)
    with fresh_canonical_aliases():
      self.assertIsNone(
          canonical_aliases.lookup_alias(
              fixture_lib.OriginalRedefinedClass,
              infer_from_attributes=True,
          )
      )
      self.assertEqual(
          canonical_aliases.lookup_alias(
              fixture_lib.OriginalRedefinedClass,
              infer_from_attributes=True,
              allow_outdated=True,
          ),
          canonical_aliases.ModuleAttributePath(
              fixture_lib.__name__,
              ("RedefinedClass",),
          ),
      )

  @parameterized.named_parameters(
      dict(
          testcase_name="ordinary_class",
          candidate=fixture_lib.DataclassWithOneChild,
          expected_result=True,
      ),
      dict(
          testcase_name="ordinary_function",
          candidate=fixture_lib.make_class_with_weird_qualname,
          expected_result=True,
      ),
      dict(
          testcase_name="moved_internal_class",
          candidate=fixture_lib.ClassDefinedInAFunction,
          alias=canonical_aliases.ModuleAttributePath(
              fixture_lib.__name__, ("ClassDefinedInAFunction",)
          ),
          expected_result=True,
      ),
      dict(
          testcase_name="confusingly_renamed_internal_class",
          candidate=fixture_lib.RedefinedClass,
          alias=canonical_aliases.ModuleAttributePath(
              fixture_lib.__name__, ("OriginalRedefinedClass",)
          ),
          expected_result=False,
      ),
      dict(
          testcase_name="private_class",
          candidate=fixture_lib._PrivateClass,
          expected_result=False,
      ),
      dict(
          testcase_name="private_function",
          candidate=fixture_lib._private_function,
          expected_result=False,
      ),
      dict(
          testcase_name="attribute_of_private_object",
          candidate=fixture_lib._PrivateClass.some_function,
          alias=canonical_aliases.ModuleAttributePath(
              fixture_lib.__name__,
              ("_PrivateClass", "some_function"),
          ),
          expected_result=False,
      ),
      dict(
          testcase_name="wrapped_functionlike_object",
          candidate=fixture_lib.wrapped_function,
          alias=canonical_aliases.ModuleAttributePath(
              fixture_lib.__name__, ("wrapped_function",)
          ),
          expected_result=True,
      ),
      dict(
          testcase_name="callable_object_but_not_wrapped_function",
          candidate=fixture_lib.some_callable_block,
          alias=canonical_aliases.ModuleAttributePath(
              fixture_lib.__name__, ("some_callable_block",)
          ),
          expected_result=False,
      ),
      dict(
          testcase_name="immutable_constant",
          candidate=fixture_lib.immutable_constant,
          alias=canonical_aliases.ModuleAttributePath(
              fixture_lib.__name__, ("immutable_constant",)
          ),
          expected_result=False,
      ),
      dict(
          testcase_name="mutable_constant",
          candidate=fixture_lib.mutable_constant,
          alias=canonical_aliases.ModuleAttributePath(
              fixture_lib.__name__, ("mutable_constant",)
          ),
          expected_result=True,
      ),
  )
  def test_default_well_known_filter(
      self, candidate, expected_result, alias=None
  ):
    with fresh_canonical_aliases():
      if alias is None:
        alias = canonical_aliases.lookup_alias(
            candidate, infer_from_attributes=True
        )
        assert alias is not None

      self.assertEqual(
          canonical_aliases.default_well_known_filter(candidate, alias),
          expected_result,
      )

  def test_local_aliases(self):
    with fresh_canonical_aliases():
      with canonical_aliases.relative_alias_names({
          "DataclassOne": fixture_lib.DataclassWithOneChild,
          "SomeOuterClass": fixture_lib.SomeOuterClass,
          "fixture_lib": fixture_lib,
      }):
        with self.subTest("class_alias"):
          self.assertEqual(
              canonical_aliases.lookup_alias(
                  fixture_lib.DataclassWithOneChild,
                  allow_relative=True,
              ),
              canonical_aliases.LocalNamePath(
                  local_name="DataclassOne", attribute_path=()
              ),
          )
        with self.subTest("parent_class_alias"):
          self.assertEqual(
              canonical_aliases.lookup_alias(
                  fixture_lib.SomeOuterClass.NestedClass,
                  allow_relative=True,
              ),
              canonical_aliases.LocalNamePath(
                  local_name="SomeOuterClass",
                  attribute_path=("NestedClass",),
              ),
          )
        with self.subTest("module_alias"):
          self.assertEqual(
              canonical_aliases.lookup_alias(
                  fixture_lib.DataclassWithTwoChildren,
                  allow_relative=True,
              ),
              canonical_aliases.LocalNamePath(
                  local_name="fixture_lib",
                  attribute_path=("DataclassWithTwoChildren",),
              ),
          )

      with canonical_aliases.relative_alias_names({
          "fixture_parent": fixture_parent,
      }):
        with self.subTest("parent_module_alias"):
          self.assertEqual(
              canonical_aliases.lookup_alias(
                  fixture_lib.DataclassWithTwoChildren,
                  allow_relative=True,
              ),
              canonical_aliases.LocalNamePath(
                  local_name="fixture_parent",
                  attribute_path=(
                      "treescope_examples_fixture",
                      "DataclassWithTwoChildren",
                  ),
              ),
          )
        with self.subTest("local_module_name_helper"):
          self.assertEqual(
              canonical_aliases.maybe_local_module_name(fixture_lib),
              "fixture_parent.treescope_examples_fixture",
          )

      with self.subTest("magic_aliases"):
        # Will detect "fixture_lib" in scope using stack inspection
        with canonical_aliases.relative_alias_names("magic"):
          self.assertEqual(
              canonical_aliases.lookup_alias(
                  fixture_lib.DataclassWithTwoChildren,
                  allow_relative=True,
              ),
              canonical_aliases.LocalNamePath(
                  local_name="fixture_lib",
                  attribute_path=("DataclassWithTwoChildren",),
              ),
          )

  @parameterized.named_parameters(
      dict(
          testcase_name="numpy.ndarray",
          target=numpy.ndarray,
          alias_string="numpy.ndarray",
      ),
      dict(
          testcase_name="jax.ShapeDtypeStruct",
          target=jax.ShapeDtypeStruct,
          alias_string="jax.ShapeDtypeStruct",
      ),
      dict(
          testcase_name="jax.vmap",
          target=jax.vmap,
          alias_string="jax.vmap",
      ),
      dict(
          testcase_name="jax.nn.relu",
          target=jax.nn.relu,
          alias_string="jax.nn.relu",
      ),
      dict(
          testcase_name="jax.nn.softmax",
          target=jax.nn.softmax,
          alias_string="jax.nn.softmax",
      ),
      dict(
          testcase_name="jax.numpy.sum",
          target=jax.numpy.sum,
          alias_string="jax.numpy.sum",
      ),
      dict(
          testcase_name="jax.scipy.special.logsumexp",
          target=jax.scipy.special.logsumexp,
          alias_string="jax.scipy.special.logsumexp",
      ),
  )
  def test_default_canonical_aliases(self, target, alias_string):
    type_registries.update_registries_for_imports()
    self.assertEqual(str(canonical_aliases.lookup_alias(target)), alias_string)


if __name__ == "__main__":
  absltest.main()
